'''
Created on 23 Oct 2020

@author: jacklok
'''

from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel
from trexlib.utils.common.cache_util import cache

import logging  
from trexlib.utils.security_util import generate_user_id, hash_password

from trexlib import conf as lib_conf
from trexmodel import conf as model_conf
import json
from json import JSONEncoder
from datetime import datetime

class FacebookSubscriber(BaseNModel, DictModel):
    subscriber_fb_id        = ndb.StringProperty(required=True)
    name                    = ndb.StringProperty(required=False)
    email                   = ndb.StringProperty(required=False)
    last_engaged_datetime   = ndb.DateTimeProperty(required=True, auto_now_add=True)
    
    @staticmethod
    def create(subscriber_fb_id, name=None, email=None): 
        fb_subscriber = FacebookSubscriber.get_by_subscriber_id(subscriber_fb_id)
        if fb_subscriber is None:
            fb_subscriber = FacebookSubscriber(subscriber_fb_id=subscriber_fb_id, name=name, email=email)
        else:
            fb_subscriber.last_engaged_datetime = datetime.now()
            fb_subscriber.name  = name
            fb_subscriber.email = email
        
        fb_subscriber.put()
        
        return fb_subscriber
        
        
    @staticmethod
    def get_by_subscriber_id(subscriber_fb_id): 
        return FacebookSubscriber.query(ndb.AND(FacebookSubscriber.subscribe_fb_id==subscriber_fb_id)).get()    

class FacebookSubscriberMessage(BaseNModel, DictModel):
    subscriber_fb_id        = ndb.StringProperty(required=True)  
    conversation_id         = ndb.StringProperty(required=True)
    
    
          



'''
Created on 13 Apr 2020

@author: jacklok
'''
from google.cloud import ndb
from trexlib.utils.common.float_util import format_float
from datetime import datetime, date, time
from trexlib.utils.string_util import is_not_empty 
from trexlib.utils.log_util import get_tracelog
from six import string_types 
from trexlib.utils.common.date_util import increase_date 
import logging
#from google.cloud.ndb._datastore_query import Cursor
from google.cloud.ndb import Cursor
from trexlib import conf as lib_conf 
from trexmodel import conf as model_conf
from google.cloud.datastore.helpers import GeoPoint
 
logger = logging.getLogger('model') 

def convert_to_serializable_value(val, none_as_empty_string=False, gmt=0, 
                                  datetime_format=lib_conf.DEFAULT_DATETIME_FORMAT, 
                                  date_format=lib_conf.DEFAULT_DATE_FORMAT, 
                                  time_format=lib_conf.DEFAULT_TIME_FORMAT):
    
    if isinstance(val, (list, tuple)):
        _list = []
        for item in val:
            _list.append(convert_to_serializable_value(item))
        value = _list
    elif isinstance(val, float):
        value = format_float(val)
    elif isinstance(val, (dict, int, bool)):
        #logger.debug('>>>>found dict, int, float, bool value=%s', val)
        value = val
    elif isinstance(val, datetime):
        gmt_datetime = increase_date(val, hour=gmt)
        value = gmt_datetime.strftime(datetime_format)
    elif isinstance(val, date):
        value = val.strftime(date_format)
    elif isinstance(val, time):
        #gmt_datetime = increase_date(val, hours=gmt)
        value = val.strftime(time_format)
    elif isinstance(val, GeoPoint):
        value = '%s,%s' % (val.latitude, val.longitude)    
    else:
        if none_as_empty_string:
            if val:
                value = str(val)
            else:
                value = ''
        else:
            if val is None:
                value = None
            else:
                value = str(val)
    #logger.debug('convert_to_serializable_value: value=%s', value)
    return value

def render_dict_value(val, full=False, show_key=True, gmt=0, 
                      datetime_format=lib_conf.DEFAULT_DATETIME_FORMAT, 
                      date_format=lib_conf.DEFAULT_DATE_FORMAT, 
                      time_format=lib_conf.DEFAULT_TIME_FORMAT):
    #logger.debug('---render_dict_value---')
    value = None
    if isinstance(val, DictModel):
        value = val.to_dict(full=full, show_key=show_key, gmt=gmt, date_format=date_format, datetime_format=datetime_format, time_format=time_format)

    elif isinstance(val, ndb.Model):
        value = val.key.urlsafe()

    elif isinstance(val, (list, tuple)):
        _list = []
        for item in val:
            _list.append(render_dict_value(item, full=full, show_key=show_key, gmt=gmt, date_format=date_format, datetime_format=datetime_format, time_format=time_format))
        value = _list

    elif isinstance(val, DictObject):
        value = val.to_dict(gmt=gmt)
    else:
        value = convert_to_serializable_value(val, gmt=gmt, date_format=date_format, datetime_format=datetime_format, time_format=time_format)
    return value

class DictObject(object):
    dict_properties = None
    
    def get_class_members(self, klass):
        ret = dir(klass)
        if hasattr(klass,'__bases__'):
            for base in klass.__bases__:
                ret = ret + self.get_class_members(base)
        return ret
    
    def uniq(self, seq ): 
        return list(set(seq))
    
    def properties(self):
        return self.__dict__

    def to_dict(self, full=True, show_key=True, dict_properties=None, gmt=lib_conf.DEFAULT_GMT, 
                        datetime_format=lib_conf.DEFAULT_DATETIME_FORMAT, 
                        date_format=lib_conf.DEFAULT_DATE_FORMAT, 
                        time_format=lib_conf.DEFAULT_TIME_FORMAT):
        logger.info('calling DictObject.to_dict')
        result = {}
        #logger.debug('properties = %s, object = %s', self.properties(), self)

        _dict_properties = dict_properties or self.dict_properties

        if _dict_properties:
            for p in _dict_properties:

                val = getattr(self, p)
                if val is not None:
                    #logger.debug('attribute=%s', p)
                    result[p] = render_dict_value(val, full=full, show_key=show_key, gmt=gmt, 
                                                  datetime_format=datetime_format, 
                                                  date_format=datetime_format, 
                                                  time_format=time_format)
        return result
    
class DictModel(ndb.Model):
    
    dict_properties             = None
    dict_full_properties        = None
    show_key_in_dict            = True

    def __get_attr_value(self, obj, attr_name):
        val = getattr(obj, attr_name)
        if isinstance(val, GeoPoint):
            return '%s,%s' % (val.latitude, val.longitude)
        
        return val

    def put(self, **kwargs):
        super(DictModel, self).put(**kwargs)

    def to_dict(self, full=False, show_key=True, dict_properties=None, gmt=lib_conf.DEFAULT_GMT, 
                        datetime_format=lib_conf.DEFAULT_DATETIME_FORMAT, 
                        date_format=lib_conf.DEFAULT_DATE_FORMAT, 
                        time_format=lib_conf.DEFAULT_TIME_FORMAT):
        result = {}
        logger.debug('DictModel: full=%s', full)
        logger.debug('DictModel: show_key=%s', show_key)
        
        if self.show_key_in_dict or show_key:
            try:
                result['key'] = self.key.urlsafe().decode("utf-8")
            except:
                pass 
        
        _dict_properties = dict_properties or self.dict_properties

        if full and self.dict_full_properties:
            _dict_properties = self.dict_full_properties

        if _dict_properties:
            for p in _dict_properties:
                if p=='key':
                    if result.get('key'):
                        continue
                    
                _p = p.split('.')

                if len(_p)>1:
                    ref_array   = _p[:-1]
                    ref_obj     = self
                    for _a in ref_array:
                        ref_obj = self.__get_attr_value(ref_obj, _a)

                    if ref_obj:
                        ref_value = self.__get_attr_value(ref_obj, _p[-1])
                else:
                    ref_value = self.__get_attr_value(self, p)

                if is_not_empty(ref_value):
                #if ref_value:
                    p = p.replace('.', '_')
                    result[p] = render_dict_value(ref_value, full=full, show_key=show_key, gmt=gmt, date_format=date_format, datetime_format=datetime_format, time_format=time_format)
                                        
        #logger.debug('DictModel: result=%s', result)
        
        return result    

class NDBModel(object):
    @classmethod
    def get_by_ndb_id(cls, ndb_id):
        if isinstance(ndb_id, string_types):
            return cls.get_by_key_name(ndb_id)
        else:
            return cls.get_by_id(ndb_id)

    def ndb_id(self):
        return self.key().id()

    def create_ndb_key(self):
        return ndb.Key(flat=self.key.flat())
    
class BaseNModel(DictModel, NDBModel):

    saved = False

    def is_saved(self):
        return self.saved

    @classmethod
    def _post_get_hook(cls, key, future):
        obj = future.get_result()
        if obj is not None:
            # test needed because post_get_hook is called even if get() fails!
            obj.saved = True

    def _post_put_hook(self, future):
        self.saved = True

    @property
    def key_in_str(self):
        return self.key.urlsafe().decode('utf-8')

    @property
    def key_name(self):
        return self.key.id()

    def equal(self, another_instance):
        if another_instance:
            return self.key_in_str == another_instance.key_in_str
        else:
            return False

    def not_equal(self, another_instance):
        if another_instance:
            return self.key_in_str != another_instance.key_in_str
        else:
            return True
    
    def delete(self):
        self.key.delete()
    
    @classmethod
    def fetch(cls, model_key):
        try:
            if model_key:
                return ndb.Key(urlsafe=model_key).get()

        except:
            logger.error('Failed to fetch entity due to %s', get_tracelog())

        return None
    
    @classmethod
    def count(cls, limit=model_conf.MAX_FETCH_RECORD): 
        query = cls.query()
        
        return cls.count_with_condition_query(query, limit=limit)
    
    @classmethod
    def count_with_condition_query(cls, query, limit=model_conf.MAX_FETCH_RECORD):    
        
        
        return query.count(limit=limit, keys_only=True)
    
    @classmethod
    def list_all(cls, offset=0,  start_cursor=None, return_with_cursor=False, keys_only=False, 
                 limit=model_conf.PAGINATION_SIZE):
        
        query = cls.query()
        
        return cls.list_all_with_condition_query(query, 
                                                 offset=offset, start_cursor=start_cursor,
                                                 return_with_cursor = return_with_cursor,
                                                 keys_only = keys_only, limit=limit
                                                 )
    
    @classmethod
    def list_all_with_condition_query(cls, query, offset=0,  start_cursor=None, return_with_cursor=False, keys_only=False, 
                 limit=model_conf.PAGINATION_SIZE):
        
        
        
        if start_cursor or return_with_cursor:
            if is_not_empty(start_cursor):
                if isinstance(start_cursor, string_types):
                    start_cursor = ndb.Cursor(urlsafe = start_cursor) if start_cursor else ndb.Cursor()
            else:
                start_cursor = ndb.Cursor()
            
            (search_results, next_cursor, more)       = query.fetch_page(page_size=limit, start_cursor=start_cursor, keys_only=keys_only)
            logger.debug('##########################################')
            
            logger.debug('list_all: search_results=%s', search_results)
            logger.debug('list_all: next_cursor=%s', next_cursor)
            logger.debug('list_all: more=%s', more) 
            
            logger.debug('##########################################')
            
            if more:
                return search_results, next_cursor.to_websafe_string().decode('utf-8')  
            else:    
                return search_results, 'None'
            
        else:
            search_results = query.fetch(limit=limit, offset=offset)
            
            return search_results
        
    def before_put(self):
        pass
    
    def after_put(self):
        pass
    
    def put(self, **kwargs):
        self.before_put()
        super(BaseNModel, self).put(**kwargs)
        self.after_put()
    
    
class FullTextSearchable(ndb.Model):
    full_text_search_field      = ndb.StringProperty(repeated=True)
    
    fulltextsearch_field_name   = None
    
    def _pre_put_hook(self):
        """before save, parse searchable field into strings"""
        _searchable_field_name = self.fulltextsearch_field_name
        logger.debug('_searchable_field_name=%s', _searchable_field_name)
        if _searchable_field_name:
            searchable_field_value = getattr(self, _searchable_field_name)
            logger.debug('searchable_field_value=%s', searchable_field_value)
            if is_not_empty(searchable_field_value):
                splitted_words = searchable_field_value.lower().split(' ')
                
                logger.debug('splitted_words=%s', splitted_words)
                
                searchable_words_list = [x for x in splitted_words if x and len(x) >= 2]
                logger.debug('searchable_words_list=%s', searchable_words_list)
                
                self.full_text_search_field = searchable_words_list
    
    
    
    @classmethod
    def full_text_count(cls, search_text_list, query=None, limit=model_conf.MAX_FETCH_RECORD_FULL_TEXT_SEARCH_PER_PAGE):
        
        logger.debug('full_text_count: search_text_list=%s', search_text_list)
        
        lowercase_words_list = []
        
        if is_not_empty(search_text_list):
            lowercase_words_list = [x.lower() for x in search_text_list]
        
            logger.debug('full_text_search: lowercase_words_list=%s', lowercase_words_list)
            
        if query is None:
            query = cls.query()
            
        
        for word in lowercase_words_list:
            query = query.filter(cls.full_text_search_field == word)
            
        logger.debug('query=%s', query)
        
        return query.count(limit=limit, keys_only=True)
            
           
        
    @classmethod
    def full_text_search(cls, text_search = None, query=None, offset=0,  start_cursor=None, return_with_cursor=False, keys_only=False, 
                         limit=model_conf.MAX_FETCH_RECORD_FULL_TEXT_SEARCH_PER_PAGE):
        if query is None:
            query = cls.query()
        
        logger.debug('##########################################')
        logger.debug('full_text_search: text_search=%s', text_search)
        logger.debug('full_text_search: query=%s', query)
        logger.debug('full_text_search: start_cursor=%s', start_cursor)
        logger.debug('full_text_search: offset=%s', offset)
        logger.debug('full_text_search: return_with_cursor=%s', return_with_cursor)
        
        logger.debug('##########################################')
        
            
        if is_not_empty(text_search):
            lowercase_words_list = [x.lower() for x in text_search]
            
            logger.debug('full_text_search: lowercase_words_list=%s', lowercase_words_list)
            
            for word in lowercase_words_list:
                query = query.filter(cls.full_text_search_field == word)
        
        if start_cursor or return_with_cursor:
            if is_not_empty(start_cursor):
                if isinstance(start_cursor, string_types):
                    start_cursor = ndb.Cursor(urlsafe = start_cursor) if start_cursor else ndb.Cursor()
            #else:
            #    start_cursor = ndb.Cursor()
            '''
            if is_not_empty(start_cursor):
                if isinstance(start_cursor, string_types):
                    start_cursor = Cursor.from_websafe_string(start_cursor)
            else:
                start_cursor = None
            '''
            if is_not_empty(start_cursor):
                logger.debug('start_cursor is empty or none')
                (search_results, next_cursor, more)       = query.fetch_page(page_size=limit, 
                                                                         start_cursor=start_cursor, 
                                                                         keys_only=keys_only)
            else:
                (search_results, next_cursor, more)       = query.fetch_page(page_size=limit, offset=offset, 
                                                                         keys_only=keys_only)
            
            logger.debug('==========================================')
            logger.debug('full_text_search: query=%s', query)
            logger.debug('full_text_search: offset=%s', offset)
            logger.debug('full_text_search: search_results count =%s', len(search_results))
            logger.debug('full_text_search: next_cursor=%s', next_cursor)
            logger.debug('full_text_search: more=%s', more) 
            
            logger.debug('==========================================')
            
            if more:
                return search_results, next_cursor.to_websafe_string().decode('utf-8')  
            else:    
                return search_results, 'None'
            
        else:
            search_results = query.fetch(limit=limit, offset=offset)
            
            logger.debug('==========================================')
            logger.debug('full_text_search: query=%s', query)
            logger.debug('full_text_search: offset=%s', offset)
            logger.debug('full_text_search: search_results count =%s', len(search_results))
            
            logger.debug('==========================================')
            
            return search_results        
            

    



'''
Created on 10 Mar 2021

@author: jacklok
'''
from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel
import trexmodel.conf as model_conf
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.merchant_models import MerchantAcct, \
    MerchantUser
import logging
from trexmodel import program_conf
from datetime import datetime

logger = logging.getLogger('model')

class VoucherBase(BaseNModel, DictModel):
    label                   = ndb.StringProperty(required=True)
    desc                    = ndb.StringProperty(required=False)
    terms_and_conditions    = ndb.TextProperty(required=False)
    completed_status        = ndb.StringProperty(required=True, choices=set(program_conf.VOUCHER_STATUS))
    configuration           = ndb.JsonProperty(required=False, default={})
    
    created_datetime        = ndb.DateTimeProperty(required=True, auto_now_add=True)
    modified_datetime       = ndb.DateTimeProperty(required=True, auto_now=True)
    published_datetime      = ndb.DateTimeProperty(required=False)
    archived_datetime       = ndb.DateTimeProperty(required=False)
    
    created_by              = ndb.KeyProperty(name="created_by", kind=MerchantUser)
    created_by_username     = ndb.StringProperty(required=False)
    
    modified_by             = ndb.KeyProperty(name="modified_by", kind=MerchantUser)
    modified_by_username    = ndb.StringProperty(required=False)
    
    published               = ndb.BooleanProperty(default=False)
    archived                = ndb.BooleanProperty(default=False)
    image_storage_filename  = ndb.StringProperty(required=False)
    image_public_url        = ndb.StringProperty(required=False)
    
    dict_properties         = ['label', 'desc', 'terms_and_conditions', 'configuration', 'created_datetime', 'modified_datetime', 'completed_status', 
                               'created_by_username', 'modified_by_username', 'is_published', 'is_archived',  
                               'image_public_url','image_storage_filename']
    
    @property
    def is_archived(self):
        return self.archived
    
    @property
    def is_published(self):
        return self.completed_status == program_conf.VOUCHER_STATUS_PUBLISH
    
class MerchantVoucher(VoucherBase):
    
    @property
    def merchant_acct(self):
        return MerchantAcct.fetch(self.key.parent().urlsafe())
    
    def to_voucher_configuration(self):
        voucher_configuration = {
                                'voucher_key'           : self.key_in_str,
                                'voucher_configuration' : self.configuration,
                                'terms_and_conditions'  : self.terms_and_conditions,
                                'image_public_url'      : self.image_public_url,
                                }
        
        return voucher_configuration
    
    @staticmethod
    def create(merchant_acct, label=None, desc=None, terms_and_conditions=None, voucher_image_url=None, created_by=None):
        created_by_username = None
        if is_not_empty(created_by):
            if isinstance(created_by, MerchantUser):
                created_by_username = created_by.username
        
        
        merchant_voucher =  MerchantVoucher(
                                        parent                  = merchant_acct.create_ndb_key(),
                                        label                   = label,
                                        desc                    = desc,
                                        terms_and_conditions    = terms_and_conditions,
                                        created_by              = created_by.create_ndb_key(),
                                        created_by_username     = created_by_username,
                                        configuration           = {},
                                        completed_status        = program_conf.VOUCHER_STATUS_BASE,
                                        image_public_url        = voucher_image_url,
                                        )
        
        merchant_voucher.put()
        return merchant_voucher
    
    @staticmethod
    def update_voucher_base_data(merchant_voucher, label=None, desc=None, terms_and_conditions=None, modified_by=None):
        
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        merchant_voucher.label                  = label
        merchant_voucher.desc                   = desc
        merchant_voucher.terms_and_conditions   = terms_and_conditions
        merchant_voucher.completed_status       = program_conf.VOUCHER_STATUS_BASE
        merchant_voucher.modified_by            = modified_by.create_ndb_key()
        merchant_voucher.modified_by_username   = modified_by_username
        
        merchant_voucher.put()
        
        return merchant_voucher
    
    @staticmethod
    def update_voucher_configuration_data(merchant_voucher, configuration=None, modified_by=None):
        
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        merchant_voucher.configuration          = configuration
        merchant_voucher.completed_status       = program_conf.VOUCHER_STATUS_CONFIGURATION
        merchant_voucher.modified_by            = modified_by.create_ndb_key()
        merchant_voucher.modified_by_username   = modified_by_username
        
        merchant_voucher.put()
        
        return merchant_voucher
    
    @staticmethod
    def update_voucher_material(merchant_voucher, image_public_url=None, image_storage_filename=None, modified_by=None):
        
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        merchant_voucher.image_public_url       = image_public_url
        merchant_voucher.image_storage_filename = image_storage_filename
        
        if merchant_voucher.completed_status!=program_conf.VOUCHER_STATUS_PUBLISH:
            merchant_voucher.completed_status       = program_conf.VOUCHER_STATUS_UPLOAD_MATERIAL 
        
        merchant_voucher.modified_by            = modified_by.create_ndb_key()
        merchant_voucher.modified_by_username   = modified_by_username
        
        merchant_voucher.put()
        
        return merchant_voucher
    
    @staticmethod
    def update_voucher_material_uploaded(merchant_voucher, modified_by=None):
        
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        merchant_voucher.completed_status       = program_conf.VOUCHER_STATUS_UPLOAD_MATERIAL
        merchant_voucher.modified_by            = modified_by.create_ndb_key()
        merchant_voucher.modified_by_username   = modified_by_username
        
        merchant_voucher.put()
        
        return merchant_voucher
    
    @staticmethod
    def publish_voucher(voucher):
        voucher.completed_status    = program_conf.VOUCHER_STATUS_PUBLISH
        voucher.apublished_datetime = datetime.now()
        voucher.put()
        
        merchant_acct = voucher.merchant_acct
        merchant_acct.update_published_voucher(voucher.to_voucher_configuration())    
        
    @staticmethod
    def archive_voucher(voucher):
        voucher.archived = True
        voucher.archived_datetime = datetime.now()
        voucher.put()
        
        merchant_acct = voucher.merchant_acct
        merchant_acct.remove_archieve_voucher(voucher.key_in_str)    
    
    @staticmethod
    def list_by_merchant_account(merchant_acct):
        return MerchantVoucher.query(ndb.AND(MerchantVoucher.archived!=True), ancestor=merchant_acct.create_ndb_key()).fetch(limit=model_conf.MAX_FETCH_RECORD)
    
    @staticmethod
    def list_archived_by_merchant_account(merchant_acct):
        return MerchantVoucher.query(ndb.AND(MerchantVoucher.archived==True), ancestor=merchant_acct.create_ndb_key()).fetch(limit=model_conf.MAX_FETCH_RECORD)
    
    @staticmethod
    def list_published_by_merchant_account(merchant_acct):
        return MerchantVoucher.query(ndb.AND(MerchantVoucher.completed_status==program_conf.VOUCHER_STATUS_PUBLISH), ancestor=merchant_acct.create_ndb_key()).fetch(limit=model_conf.MAX_FETCH_RECORD)

class BrandVouhcer(VoucherBase):
    #merchant_acct           = ndb.KeyProperty(name="merchant_acct", kind=MerchantAcct)
    pass



'''
Created on 19 Feb 2021

@author: jacklok
'''
from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel, FullTextSearchable
from trexmodel.models.datastore.system_models import SentEmail
from trexmodel.models.datastore.user_models import UserMin
import trexmodel.conf as model_conf
from trexlib.utils.security_util import generate_user_id, hash_password
from trexlib.utils.string_util import random_number
import logging
from datetime import datetime, timedelta
from trexlib.utils.common.date_util import parse_datetime
from trexmodel import conf
from google.auth._default import default
from trexmodel.models.datastore.system_models import Tagging


logger = logging.getLogger('model')

class CorporateBase(BaseNModel, DictModel, FullTextSearchable):
    
    company_name            = ndb.StringProperty(required=True)
    contact_name            = ndb.StringProperty(required=False)
    address                 = ndb.StringProperty(required=False)
    office_phone            = ndb.StringProperty(required=False)
    mobile_phone            = ndb.StringProperty(required=False)
    fax_phone               = ndb.StringProperty(required=False)
    email                   = ndb.StringProperty(required=False)
    country                 = ndb.StringProperty(required=False, default='my')
    status                  = ndb.StringProperty(required=False)
    
    modified_datetime       = ndb.DateTimeProperty(required=True, auto_now=True)
    registered_datetime     = ndb.DateTimeProperty(required=True, auto_now_add=True)
    plan_start_date         = ndb.DateProperty(required=True)
    plan_end_date           = ndb.DateProperty(required=True)
    
    fulltextsearch_field_name   = 'company_name'
    
    

class CorporateAcct(CorporateBase):
    account_code                = ndb.StringProperty(required=False)
    logo_public_url             = ndb.StringProperty(required=False)
    logo_storage_filename       = ndb.StringProperty(required=False)
    dashboard_stat_figure       = ndb.JsonProperty()
    currency_code               = ndb.StringProperty(required=False, default='myr')
    
    
    


'''
Created on 27 May 2020

@author: jacklok
'''

from google.cloud import datastore
import logging


def read(client, entity_key):
    # [START datastore_basic_entity]
    entity = datastore.Entity(client.key(entity_key))
    
    logging.debug('entity=%s', entity)
    
    return entity




'''
Created on 20 Apr 2020

@author: jacklok
'''
from trexmodel import conf as model_conf
from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel
 

class Country(BaseNModel, DictModel):
    cnty_id                 = ndb.IntegerProperty(required=True)
    cnty_code               = ndb.StringProperty(required=True)
    cnty_name               = ndb.StringProperty(required=True)
    locale                  = ndb.StringProperty(required=True)
    gmt                     = ndb.StringProperty(required=True)
    trunk_code              = ndb.StringProperty(required=True)
    
    
class ContactUs(BaseNModel, DictModel):
    contact_name                = ndb.StringProperty(required=True)
    contact_email               = ndb.StringProperty(required=True)
    contact_subject             = ndb.StringProperty(required=False)
    contact_message             = ndb.TextProperty(required=True)
    created_datetime            = ndb.DateTimeProperty(required=True, auto_now=True)
    
    dict_properties = ['contact_name', 'contact_email', 'contact_subject', 'contact_message', 
                       'created_datetime']
    
    @staticmethod
    def create(contact_name, contact_email=None, contact_subject=None, contact_message=None):
        ContactUs(
                contact_name    = contact_name,
                contact_email   = contact_email,
                contact_subject = contact_subject,
                contact_message = contact_message,
                ).put()
                
    @staticmethod
    def list(offset=0, limit=10):
        return ContactUs.query().order(-ContactUs.created_datetime).fetch(offset=offset, limit=limit)    
                
class Feedback(BaseNModel, DictModel):
    name                    = ndb.StringProperty(required=True)
    email                   = ndb.StringProperty(required=True)
    rating                  = ndb.StringProperty(required=True)
    message                 = ndb.TextProperty(required=True)
    created_datetime        = ndb.DateTimeProperty(required=True, auto_now=True)
    
    @staticmethod
    def create(name=None, email=None, rating=None, message=None):
        ContactUs(
                name        = name,
                email       = email,
                rating      = rating,
                message     = message,
                ).put()                
        
        
class SentEmail(BaseNModel, DictModel):
    email_id                = ndb.StringProperty(required=True)
    batch_id                = ndb.StringProperty(required=True)
    to                      = ndb.StringProperty(required=True)
    subject                 = ndb.StringProperty(required=True)
    html                    = ndb.StringProperty(required=False) 
    body                    = ndb.StringProperty(required=False)
    sent_datetime           = ndb.DateTimeProperty(required=True, auto_now=True)
    sent_response           = ndb.StringProperty(required=False)
    sent_status             = ndb.BooleanProperty(default=False)
    click_status            = ndb.BooleanProperty(default=False)
    read_status             = ndb.BooleanProperty(default=False)
    open_status             = ndb.BooleanProperty(default=False)
    bounce_status           = ndb.BooleanProperty(default=False)
    spam_status             = ndb.BooleanProperty(default=False)
    block_status            = ndb.BooleanProperty(default=False)
    
    
    @classmethod
    def get_by_email_id(cls, email_id):
        return cls.query(ndb.AND(cls.email_id==email_id)).get()
    
    @classmethod
    def update_sent_status_by_email_id(cls, email_id, sent_status=True):
        sent_email = cls.get_by_email_id(email_id)
        if sent_email:
            sent_email.sent_status = sent_status
            sent_email.put()
            
    @classmethod
    def update_open_status_by_email_id(cls, email_id, open_status=True):
        sent_email = cls.get_by_email_id(email_id)
        if sent_email:
            sent_email.open_status = open_status
            sent_email.put()
            
    @classmethod
    def update_click_status_by_email_id(cls, email_id, click_status=True):
        sent_email = cls.get_by_email_id(email_id)
        if sent_email:
            sent_email.click_status = click_status
            sent_email.put()  
            
    @classmethod
    def update_read_status_by_email_id(cls, email_id, read_status=True):
        sent_email = cls.get_by_email_id(email_id)
        if sent_email:
            sent_email.read_status = read_status
            sent_email.put()          
            
    @classmethod
    def update_bounce_status_by_email_id(cls, email_id, bounce_status=True):
        sent_email = cls.get_by_email_id(email_id)
        if sent_email:
            sent_email.bounce_status = bounce_status
            sent_email.put()  
            
    @classmethod
    def update_spam_status_by_email_id(cls, email_id, spam_status=True):
        sent_email = cls.get_by_email_id(email_id)
        if sent_email:
            sent_email.spam_status = spam_status
            sent_email.put() 
            
    @classmethod
    def update_block_status_by_email_id(cls, email_id, block_status=True):
        sent_email = cls.get_by_email_id(email_id)
        if sent_email:
            sent_email.block_status = block_status
            sent_email.put()                                        
    
    @classmethod
    def count_by_batch_id(cls, batch_id, limit=model_conf.MAX_FETCH_RECORD):
        return cls.query(ndb.AND(cls.batch_id==batch_id)).count(limit=limit)
    
    @classmethod
    def list_by_batch_id(cls, batch_id, offset=0, limit=model_conf.MAX_FETCH_RECORD):
        return cls.query(ndb.AND(cls.batch_id==batch_id)).fetch(offset=offset, limit=limit)
    
class Tagging(BaseNModel, DictModel):    
    label   = ndb.StringProperty(required=True)
    desc    = ndb.StringProperty(required=False)
    created_datetime    = ndb.DateTimeProperty(required=True, auto_now_add=True)
    
    dict_properties = ['label', 'desc', 'created_datetime']
    
    @classmethod
    def create_tag(cls,parent=None, label=None, desc=None):
        created_tag = cls(
                            parent  = parent,
                            label   = label,
                            desc    = desc,
                            )
        
        created_tag.put()
        
        return created_tag
    
    @classmethod
    def get_by_label(cls, parent, label):
        return cls.query(ndb.AND(cls.label==label), ancestor=parent).fetch(limit=1)
    
        
        
    
    
    


'''
Created on 27 Jan 2021

@author: jacklok
'''
import functools
from google.cloud import ndb
import logging

import time
logger = logging.getLogger('model')

def model_transactional(
        retries = 1,
        read_only=False,
        join=True,
        xg=True,
        propagation=None,
        desc=None):
    
    if desc is None:
        desc=''
    
    
    def transactional_wrapper(wrapped):
        @functools.wraps(wrapped)
        def transactional_inner_wrapper(*args, **kwargs):
            @ndb.transactional(retries=retries, read_only=read_only, join=join, xg=xg, propagation=propagation)
            def callback():
                start = time.time()
                try:
                    result =  wrapped(*args, **kwargs)
                except:
                    pass
                end = time.time()
                total_nanoseconds = int((end-start) * 1000000)
                logger.debug('-------------------{}--------------------------'.format(desc))
                logger.debug('transaction took %s nanoseconds', total_nanoseconds)
                logger.debug('----------------------------------------------------------')
                return result
            
            return callback()

        return transactional_inner_wrapper

    return transactional_wrapper 
    
        



'''
Created on 7 Apr 2021

@author: jacklok
'''

from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel
import trexmodel.conf as model_conf
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.merchant_models import MerchantAcct, \
    MerchantUser
import logging

from datetime import datetime

logger = logging.getLogger('model')

class MembershipBase(BaseNModel, DictModel):
    '''
    Merchant Acct as ancestor
    
    '''
    
    label                   = ndb.StringProperty(required=True)
    desc                    = ndb.StringProperty(required=False)
    
    expiration_type         = ndb.StringProperty(required=True)
    expiration_value        = ndb.IntegerProperty(required=False)
    
    archived                = ndb.BooleanProperty(default=False)
    
    created_datetime        = ndb.DateTimeProperty(required=True, auto_now_add=True)
    modified_datetime       = ndb.DateTimeProperty(required=True, auto_now=True)
    archived_datetime       = ndb.DateTimeProperty(required=False)
    
    created_by              = ndb.KeyProperty(name="created_by", kind=MerchantUser)
    created_by_username     = ndb.StringProperty(required=False)
    
    modified_by             = ndb.KeyProperty(name="modified_by", kind=MerchantUser)
    modified_by_username    = ndb.StringProperty(required=False)

    dict_properties         = ['label', 'desc', 'expiration_type', 'expiration_value', 'created_datetime', 'modified_datetime']
    
    @property
    def merchant_acct(self):
        return MerchantAcct.fetch(self.key.parent().urlsafe())
    
    @classmethod
    def list_by_merchant_acct(cls, merchant_acct, is_archived=False):
        return cls.query(ndb.AND(MerchantMembership.archived == is_archived), ancestor=merchant_acct.create_ndb_key()).fetch(limit=model_conf.MAX_FETCH_RECORD)
    
    @classmethod
    def create(cls, merchant_acct, label, desc=None, expiration_type=None, expiration_value=None, 
               created_by=None):
        
        created_by_username = None
        if is_not_empty(created_by):
            if isinstance(created_by, MerchantUser):
                created_by_username = created_by.username
        
        merchant_membership = cls(
                                                parent                  = merchant_acct.create_ndb_key(),
                                                label                   = label,
                                                desc                    = desc,
                                                expiration_type         = expiration_type,
                                                expiration_value        = expiration_value,
                                                created_by              = created_by.create_ndb_key(),
                                                created_by_username     = created_by_username,
                                                )
        
        merchant_membership.put()
        
        return merchant_membership
    
    @classmethod
    def update(cls, merchant_membership, label, desc=None, expiration_type=None, expiration_value=None, 
               modified_by=None):
        modified_by_username = None
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        merchant_membership.label                   = label
        merchant_membership.desc                    = desc
        merchant_membership.expiration_type         = expiration_type
        merchant_membership.expiration_value        = expiration_value
        merchant_membership.modified_by             = modified_by.create_ndb_key()
        merchant_membership.modified_by_username    = modified_by_username
        
        merchant_membership.put()

class MerchantMembership(MembershipBase):
    
    
    def to_configuration(self):
        membership_configuration = {
                                    'membership_key'                : self.key_in_str,
                                    'label'                         : self.label,
                                    'expiration_type'               : self.expiration_type,
                                    'expiration_value'              : self.expiration_value,
                                    }
        
        return membership_configuration
    
    @staticmethod
    def archive_membership(membership):
        membership.archived = True
        membership.archived_datetime = datetime.now()
        membership.put()
        
        merchant_acct = membership.merchant_acct
        merchant_acct.remove_archieve_basic_membership(membership.key_in_str)    
        
    


class MerchantTierMembership(MembershipBase):        
    entitle_qualification_type          = ndb.StringProperty(required=True)
    entitle_qualification_value         = ndb.FloatProperty(required=False)
    upgrade_expiry_type                 = ndb.StringProperty(required=True)
    
    dict_properties         = ['label', 'desc', 'expiration_type', 'expiration_value', 'entitle_qualification_type', 
                               'entitle_qualification_value', 'upgrade_expiry_type', 'created_datetime', 'modified_datetime'
                               ]
    
    
    def to_configuration(self):
        membership_configuration = {
                                    'membership_key'                : self.key_in_str,
                                    'label'                         : self.label,
                                    'expiration_type'               : self.expiration_type,
                                    'expiration_value'              : self.expiration_value,
                                    'entitle_qualification_type'    : self.entitle_qualification_type,
                                    'entitle_qualification_value'   : self.entitle_qualification_value,
                                    'upgrade_expiry_type'           : self.upgrade_expiry_type,
                                    }
        
        return membership_configuration
        
    @staticmethod
    def create(merchant_acct, label=None, desc=None, expiration_type=None, expiration_value=None, 
               entitle_qualification_type=None, entitle_qualification_value=None, upgrade_expiry_type=None,
               created_by=None):
        
        created_by_username = None
        if is_not_empty(created_by):
            if isinstance(created_by, MerchantUser):
                created_by_username = created_by.username
        
        merchant_membership = MerchantTierMembership(
                                                parent                          = merchant_acct.create_ndb_key(),
                                                label                           = label,
                                                desc                            = desc,
                                                expiration_type                 = expiration_type,
                                                expiration_value                = expiration_value,
                                                entitle_qualification_type      = entitle_qualification_type,
                                                entitle_qualification_value     = entitle_qualification_value,
                                                upgrade_expiry_type             = upgrade_expiry_type,
                                                created_by                      = created_by.create_ndb_key(),
                                                created_by_username             = created_by_username,
                                                )
        
        merchant_membership.put()
        
        membership_key = merchant_membership.key_in_str
        
        logger.debug('membership_key=%s', membership_key)
        
        merchant_acct.add_tier_membership(merchant_membership.to_configuration())
        
        return merchant_membership
    
    @staticmethod
    def update(merchant_membership, label=None, desc=None, expiration_type=None, expiration_value=None, 
               entitle_qualification_type=None, entitle_qualification_value=None, upgrade_expiry_type=None,
               modified_by=None):
        
        modified_by_username = None
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        merchant_membership.label                           = label
        merchant_membership.desc                            = desc
        merchant_membership.expiration_type                 = expiration_type
        merchant_membership.expiration_value                = expiration_value
        merchant_membership.entitle_qualification_type      = entitle_qualification_type
        merchant_membership.entitle_qualification_value     = entitle_qualification_value
        merchant_membership.entitle_qualification_value     = entitle_qualification_value
        merchant_membership.upgrade_expiry_type             = upgrade_expiry_type
        merchant_membership.modified_by_username            = modified_by_username
        
        merchant_membership.put()
        
        merchant_acct = merchant_membership.merchant_acct
        
        merchant_acct.update_tier_membership(merchant_membership.to_configuration())
        
        return merchant_membership
    
    @staticmethod
    def archive_membership(membership):
        membership.archived = True
        membership.archived_datetime = datetime.now()
        membership.put()
        
        merchant_acct = membership.merchant_acct
        merchant_acct.remove_archieve_tier_membership(membership.key_in_str)   
            


'''
Created on 1 Apr 2021

@author: jacklok
'''

from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel
from trexmodel.models.datastore.user_models import User
from trexmodel.models.datastore.customer_models import Customer
import trexmodel.conf as model_conf
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet,\
    MerchantUser
import logging
from trexmodel import conf
from datetime import datetime,date, timedelta
from trexmodel.conf import HANDLE_DATETIME_WITH_GMT, SERVER_DATETIME_GMT
from trexmodel.models.datastore.reward_models import CustomerPointReward,\
    CustomerStampReward, CustomerEntitledVoucher
from trexmodel.models.datastore.customer_model_helpers import update_reward_summary_with_reverted_reward
from trexmodel.utils.model.model_util import generate_transaction_id

logger = logging.getLogger('model')


class CustomerTransaction(BaseNModel, DictModel):
    '''
    
    '''
    transact_merchant            = ndb.KeyProperty(name="transact_merchant", kind=MerchantAcct)
    transact_outlet             = ndb.KeyProperty(name="transact_outlet", kind=Outlet)
    
    transact_datetime           = ndb.DateTimeProperty(required=True)
    created_datetime            = ndb.DateTimeProperty(required=True, auto_now=True)
    
    transaction_id              = ndb.StringProperty(required=True)
    invoice_id                  = ndb.StringProperty(required=False)
    remarks                     = ndb.StringProperty(required=False)
    
    tax_amount                  = ndb.FloatProperty(required=False, default=.0)
    transact_amount             = ndb.FloatProperty(required=True)
    
    transact_by                 = ndb.KeyProperty(name="created_by", kind=MerchantUser)
    transact_by_username        = ndb.StringProperty(required=False)
    
    entitled_reward_summary     = ndb.JsonProperty()
    entitled_voucher_summary    = ndb.JsonProperty()
    
    is_revert                   = ndb.BooleanProperty(required=False, default=False)
    reverted_datetime           = ndb.DateTimeProperty(required=False)
    reverted_by                 = ndb.KeyProperty(name="reverted_by", kind=MerchantUser)
    reverted_by_username        = ndb.StringProperty(required=False)
    
    
    dict_properties         = ['transaction_id', 'invoice_id', 'remarks', 'tax_amount', 'transact_amount', 
                               'entitled_reward_summary', 'entitled_voucher_summary', 'transact_customer_acct', 'transact_outlet_details',
                               'transact_datetime', 'created_datetime',  'transact_outlet_key', 'is_revert', 'reverted_datetime',
                               'transact_by_username']
    
    def to_transaction_details_json(self):
        pass
    
    @property
    def transact_customer_acct(self):
        return Customer.fetch(self.key.parent().urlsafe())
    
    @property
    def transact_user_acct_key(self):
        return Customer.fetch(self.key.parent().urlsafe()).registered_user_acct_key
    
    @property
    def transact_merchant_acct(self):
        return MerchantAcct.fetch(self.transact_merchant.urlsafe())
    
    @property
    def transact_outlet_key(self):
        return self.transact_outlet.urlsafe()
    
    @property
    def transact_merchant_acct_key(self):
        return self.transact_merchant.urlsafe()
    
    @property
    def transact_customer_key(self):
        return self.key.parent().urlsafe()
    
    @property
    def transact_outlet_details(self):
        return Outlet.fetch(self.transact_outlet.urlsafe())
    
    @property
    def transact_by_user(self):
        return MerchantUser.fetch(self.transact_by.urlsafe())
    
    @staticmethod
    def create(customer, transact_amount=.0, tax_amount=.0, invoice_id=None, remarks=None, transact_outlet=None, transact_by=None, transact_datetime=None):
        
        if is_not_empty(transact_by):
            if isinstance(transact_by, MerchantUser):
                transact_by_username = transact_by.username

        
        transaction_id = generate_transaction_id()
        
        if transact_datetime is None:
            transact_datetime = datetime.now()
            if HANDLE_DATETIME_WITH_GMT:
                transact_datetime = transact_datetime - timedelta(hours=int(SERVER_DATETIME_GMT))
        
        logger.debug('generated transaction_id=%s', transaction_id)
        logger.debug('invoice_id=%s', invoice_id)
        logger.debug('tax_amount=%s', tax_amount)
        logger.debug('transact_amount=%s', transact_amount)
        logger.debug('transact_datetime=%s', transact_datetime)
        logger.debug('transact_by_username=%s', transact_by_username)
        
        customer.last_transact_datetime = transact_datetime
        
        customer_transaction = CustomerTransaction(
                                                    parent                  = customer.create_ndb_key(),
                                                    
                                                    transact_merchant       = customer.registered_merchant_acct.create_ndb_key(),
                                                    transact_outlet         = transact_outlet.create_ndb_key(),
                                                    
                                                    tax_amount              = tax_amount,
                                                    transact_amount         = transact_amount,
                                                    
                                                    transaction_id          = transaction_id,
                                                    invoice_id              = invoice_id,
                                                    remarks                 = remarks,
                                                    
                                                    transact_by             = transact_by.create_ndb_key(),
                                                    transact_by_username    = transact_by_username,
                                                    
                                                    transact_datetime       = transact_datetime,
                                                    )
        
        customer_transaction.put()
        customer.put()
        
        return customer_transaction
    
    def revert(self, customer_acct, reverted_by):
        transaction_id = self.transaction_id
        
        entitled_point_details_list     = CustomerPointReward.list_by_transaction_id(transaction_id)
        
        entitled_stamp_details_list     = CustomerStampReward.list_by_transaction_id(transaction_id)
        
        entiteld_vouchers_list          = CustomerEntitledVoucher.list_by_transaction_id(transaction_id)
        
        is_transaction_reward_used      = False
        
        customer_reward_summary = customer_acct.reward_summary
        
        for p in entitled_point_details_list:
            if p.is_used:
                is_transaction_reward_used = True
                break
        
        if is_transaction_reward_used is False:
            for p in entitled_stamp_details_list:
                if p.is_used:
                    is_transaction_reward_used = True
                    break
                
        if is_transaction_reward_used is False:
            for p in entiteld_vouchers_list:
                if p.is_used:
                    is_transaction_reward_used = True
                    break
        
        if is_transaction_reward_used is False:
            reverted_datetime               = datetime.now()
            self.is_revert                  = True
            self.reveted_datetime           = reverted_datetime
            self.reverted_by                = reverted_by.create_ndb_key()
            self.reverted_by_username       = reverted_by.username
            self.put()
            
            for p in entitled_point_details_list:
                customer_reward_summary = update_reward_summary_with_reverted_reward(customer_reward_summary, p)
                
            for p in entitled_stamp_details_list:
                customer_reward_summary = update_reward_summary_with_reverted_reward(customer_reward_summary, p)
                
            for p in entiteld_vouchers_list:
                p.is_reverted = True    
            
            customer_acct.reward_summary = customer_reward_summary
            customer_acct.put()
            
            return True
        
        else:
            return False
    
    @staticmethod
    def list_customer_transaction(customer_acct, offset=0, limit=conf.PAGINATION_SIZE, start_cursor=None, return_with_cursor=False):
        query = CustomerTransaction.query(ancestor = customer_acct.create_ndb_key()).order(-CustomerTransaction.transact_datetime)
        
        return CustomerTransaction.list_all_with_condition_query(query, offset=offset, limit=limit, start_cursor=start_cursor, return_with_cursor=return_with_cursor)
    
    @staticmethod
    def get_by_transaction_id(transaction_id):
        return CustomerTransaction.query(CustomerTransaction.transaction_id==transaction_id).get()
    
    @staticmethod
    def list_transaction_by_date(transact_date, transact_outlet=None, offset=0, limit=conf.PAGINATION_SIZE, start_cursor=None, return_with_cursor=False):
        
        transact_datetime           = datetime.combine(transact_date, datetime.min.time())
        next_day_transact_datetime  = transact_datetime + timedelta(days=1)
        
        logger.debug('transact_datetime=%s',transact_datetime)
        logger.debug('next_day_transact_datetime=%s',next_day_transact_datetime)
        
        if transact_outlet:
            query = CustomerTransaction.query(ndb.AND(
                                    CustomerTransaction.transact_datetime  >= transact_datetime,
                                    CustomerTransaction.transact_datetime  <  next_day_transact_datetime,
                                    CustomerTransaction.transact_outlet == transact_outlet.create_ndb_key()
                                    )).order(-CustomerTransaction.transact_datetime)
        else:
            query = CustomerTransaction.query(ndb.AND(
                                    CustomerTransaction.transact_datetime  >= transact_datetime,
                                    CustomerTransaction.transact_datetime  <  next_day_transact_datetime,
                                    )).order(-CustomerTransaction.transact_datetime)
        
        return CustomerTransaction.list_all_with_condition_query(query, offset=offset, limit=limit, start_cursor=start_cursor, return_with_cursor=return_with_cursor)
    
    @staticmethod
    def list_all(offset=0, limit=conf.MAX_FETCH_RECORD):
        
        query = CustomerTransaction.query()
        
        return CustomerTransaction.list_all_with_condition_query(query, offset=offset, limit=limit)
    
    @staticmethod
    def count_customer_transaction(customer_acct, limit=conf.MAX_FETCH_RECORD):
        query = CustomerTransaction.query(ancestor = customer_acct.create_ndb_key())
        
        return CustomerTransaction.count_with_condition_query(query, limit=limit)
    
    @staticmethod
    def count_transaction_by_date(transact_date, transact_outlet=None, limit=conf.MAX_FETCH_RECORD):
        
        transact_datetime           = datetime.combine(transact_date, datetime.min.time())
        next_day_transact_datetime  = transact_datetime + timedelta(days=1)
        
        logger.debug('transact_datetime=%s',transact_datetime)
        logger.debug('next_day_transact_datetime=%s',next_day_transact_datetime)
        
        if transact_outlet:
            query = CustomerTransaction.query(ndb.AND(
                                    CustomerTransaction.transact_datetime  >= transact_datetime,
                                    CustomerTransaction.transact_datetime  <  next_day_transact_datetime,
                                    CustomerTransaction.transact_outlet == transact_outlet.create_ndb_key()
                                    ))
        else:
            query = CustomerTransaction.query(ndb.AND(
                                    CustomerTransaction.transact_datetime  >= transact_datetime,
                                    CustomerTransaction.transact_datetime  <  next_day_transact_datetime,
                                    ))
        
        return CustomerTransaction.count_with_condition_query(query, limit=limit)
    
class CustomerTransactionWithRewardDetails(object):    
    
    def __init__(self, transaction_details, reward_details):
        self.transact_customer_key          = transaction_details.transact_customer_key
        self.transact_merchant_acct_key     = transaction_details.transact_merchant_acct_key
        self.transact_outlet_key            = transaction_details.transact_outlet_key
        self.transact_datetime              = transaction_details.transact_datetime
        self.transaction_id                 = transaction_details.transaction_id
        self.transact_amount                = transaction_details.transact_amount
        
        self.reward_format                  = reward_details.reward_format
        self.reward_amount                  = reward_details.reward_amount
        self.expiry_date                    = reward_details.expiry_date
        self.rewarded_datetime              = reward_details.rewarded_datetime
        self.reward_format_key              = reward_details.reward_format_key
        
    
    
    
    
        
      
        
        
    
        


'''
Created on 15 Apr 2020

@author: jacklok
'''
from google.cloud import ndb
from google.oauth2 import service_account
from google.auth import crypt
import io, json, logging
from trexmodel import conf as model_conf
import six
from datetime import datetime
from trexlib.utils.string_util import random_number

logger = logging.getLogger('root')

def create_db_client(info=None, credential_filepath=None, namespace=None, caller_info=None):
    
    #datastore_cred = service_account.Credentials.from_service_account_file(lib_conf.DATASTORE_CREDENTIAL_PATH)
    logger.debug('create_db_client: caller_info=%s', caller_info)
    #logger.debug('create_db_client: info=%s', info)
    if info:
        datastore_cred = service_account.Credentials.from_service_account_info(info)
    else: 
        if credential_filepath:
            datastore_cred = service_account.Credentials.from_service_account_file(credential_filepath)
        else:
            datastore_cred = service_account.Credentials.from_service_account_file(
                                                            model_conf.DATASTORE_CREDENTIAL_PATH)
            
    client = ndb.Client(credentials=datastore_cred, project=model_conf.MODEL_PROJECT_ID, namespace=namespace)
    
    return client


def from_dict(data, require=None):
    
    keys_needed = set(require if require is not None else [])

    missing = keys_needed.difference(six.iterkeys(data))

    if missing:
        raise ValueError(
            "Service account info was not in the expected format, missing "
            "fields {}.".format(", ".join(missing))
        )

    # Create a signer.
    signer = crypt.RSASigner.from_service_account_info(data)

    return signer

def read_service_account_file(credential_filepath=model_conf.DATASTORE_CREDENTIAL_PATH):
    with io.open(credential_filepath, "r", encoding="utf-8") as json_file:
        data = json.load(json_file)
        return data, from_dict(data)

def generate_transaction_id(prefix=''):
    now                 = datetime.now()
    datetime_str        = now.strftime('%y%m%d%H%M%S')
    random_str_value    = random_number(6)
    
    return prefix[0:7] + datetime_str + random_str_value
    
    




'''
Created on 15 May 2020

@author: jacklok
'''

from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel, FullTextSearchable
from trexmodel.models.datastore.system_models import SentEmail
from trexmodel.models.datastore.user_models import UserMin
import trexmodel.conf as model_conf
from trexlib.utils.security_util import generate_user_id, hash_password
from trexlib.utils.string_util import random_number, is_empty, is_not_empty,\
    split_by_length
import logging
from datetime import datetime, timedelta
from trexlib.utils.common.date_util import parse_datetime
from trexmodel import conf
from google.auth._default import default
from trexmodel.models.datastore.system_models import Tagging

logger = logging.getLogger('model')

class MerchantMin(BaseNModel, DictModel, FullTextSearchable):
    
    company_name                = ndb.StringProperty(required=True)
    contact_name                = ndb.StringProperty(required=False)
    address                     = ndb.StringProperty(required=False)
    office_phone                = ndb.StringProperty(required=False)
    mobile_phone                = ndb.StringProperty(required=False)
    fax_phone                   = ndb.StringProperty(required=False)
    email                       = ndb.StringProperty(required=False)
    country                     = ndb.StringProperty(required=False, default='my')
    status                      = ndb.StringProperty(required=False)
    
    modified_datetime           = ndb.DateTimeProperty(required=True, auto_now=True)
    registered_datetime         = ndb.DateTimeProperty(required=True, auto_now_add=True)
    plan_start_date             = ndb.DateProperty(required=True)
    plan_end_date               = ndb.DateProperty(required=True)
    
    fulltextsearch_field_name   = 'company_name'
    
    
    @property
    def gmt_hour(self):
        return conf.DEFAULT_GMT_HOURS
    

class MerchantAcct(MerchantMin):
    account_code                            = ndb.StringProperty(required=False)
    logo_public_url                         = ndb.StringProperty(required=False)
    logo_storage_filename                   = ndb.StringProperty(required=False)
    dashboard_stat_figure                   = ndb.JsonProperty()
    currency_code                           = ndb.StringProperty(required=False, default='myr')
    
    published_program_configuration         = ndb.JsonProperty()
    published_voucher_configuration         = ndb.JsonProperty()
    membership_configuration                = ndb.JsonProperty()
    tier_membership_configuration           = ndb.JsonProperty()
    
    stat_figure_update_interval_in_minutes  = conf.MERCHANT_STAT_FIGURE_UPDATE_INTERVAL_IN_MINUTES
    stat_figure_update_datetime_format      = '%d-%m-%Y %H:%M:%S'
    
    dict_properties = ['company_name', 'contact_name', 'mobile_phone', 'office_phone', 'fax_phone', 'email', 'account_code', 'country',
                       'registered_datetime', 'modified_datetime', 'plan_start_date', 'plan_end_date', 'currency_code', 
                       'published_program_configuration', 'published_voucher_configuration', 'membership_configuration', 
                       'tier_membership_configuration']
    
    @staticmethod
    def format_account_code(account_code):
        if is_not_empty(account_code):
            if len(account_code) == 16:
                account_code = '-'.join(split_by_length(account_code,4))
        return account_code    
    
    @staticmethod
    def search_merchant_account(company_name=None, account_code=None,
                                 offset=0, start_cursor=None, limit=model_conf.MAX_FETCH_RECORD):
        
        search_text_list = None
        query = MerchantAcct.query()
        
        if is_not_empty(company_name):
            search_text_list = company_name.split(' ')
            
        elif is_not_empty(account_code):
            query = query.filter(MerchantAcct.account_code==account_code)
            
        
        
        total_count                         = MerchantAcct.full_text_count(search_text_list, query, conf.MAX_FETCH_RECORD_FULL_TEXT_SEARCH)
        
        (search_results, next_cursor)       = MerchantAcct.full_text_search(search_text_list, query, offset=offset, 
                                                                   start_cursor=start_cursor, return_with_cursor=True, 
                                                                   limit=limit)
        
        return (search_results, total_count, next_cursor)
    
    
    def flush_dirty_program_configuration(self):
        if self.published_program_configuration and len(self.published_program_configuration)>0:
            existing_programs_list  = self.published_program_configuration.get('programs')
            new_programs_list       = []
            for p in existing_programs_list:
                try:
                    program = ndb.Key(urlsafe=p.get('program_key')).get()
                    if program.archived is False:
                        new_programs_list.append(p)
                except:
                    pass
            
            self.published_program_configuration['programs']    = new_programs_list
            self.published_program_configuration['count']       = len(new_programs_list)
        else:
            self.published_program_configuration = {'programs':[], 'count':0}
            
        self.put()
        
    
    def flush_dirty_membership_configuration(self):    
        if self.membership_configuration and len(self.membership_configuration.get('memberships'))>0:
            existing_memberships_list  = self.membership_configuration.get('memberships')
            new_memberships_list       = []
            for p in existing_memberships_list:
                try:
                    membership = ndb.Key(urlsafe=p.get('membership_key')).get()
                    if membership.archived is False:
                        new_memberships_list.append(p)
                except:
                    pass
            
            self.membership_configuration['memberships'] = new_memberships_list
            self.membership_configuration['count']       = len(new_memberships_list)
            
        else:
            self.membership_configuration = {'memberships':[], 'count':0}
        
        self.put()
        
    def flush_dirty_tier_membership_configuration(self):    
        if self.tier_membership_configuration and len(self.tier_membership_configuration.get('memberships'))>0:
            existing_memberships_list  = self.tier_membership_configuration.get('memberships')
            new_memberships_list       = []
            for p in existing_memberships_list:
                try:
                    membership = ndb.Key(urlsafe=p.get('membership_key')).get()
                    if membership.archived is False:
                        new_memberships_list.append(p)
                except:
                    pass
            
            self.tier_membership_configuration['memberships'] = new_memberships_list
            self.tier_membership_configuration['count']       = len(new_memberships_list)
            
        else:
            self.tier_membership_configuration = {'memberships':[], 'count':0}
        
        self.put()
        
    def flush_and_update_membership_configuration(self, memberships_list):    
        
        logger.debug('flush_and_update_membership_configuration: memberships_list count=%d ', len(memberships_list))
        
        membership_configuration_list = []
        
        for m in memberships_list:
            membership_configuration_list.append(m.to_configuration())
            
        self.membership_configuration = {'memberships':membership_configuration_list, 'count':len(membership_configuration_list)}
        
        self.put()
        
    def flush_and_update_tier_membership_configuration(self, memberships_list):
        
        logger.debug('flush_and_update_tier_membership_configuration: memberships_list count=%d ', len(memberships_list))  
        
        membership_configuration_list = []
        
        for m in memberships_list:
            membership_configuration_list.append(m.to_configuration())
            
        self.tier_membership_configuration = {'memberships':membership_configuration_list, 'count':len(membership_configuration_list)}
        
        self.put()                
        
    def flush_dirty_voucher_configuration(self):
        if self.published_voucher_configuration and len(self.published_voucher_configuration)>0:
            existing_vouchers_list  = self.published_voucher_configuration.get('vouchers')
            new_vouchers_list       = []
            for p in existing_vouchers_list:
                try:
                    program = ndb.Key(urlsafe=p.get('program_key')).get()
                    if program.archived is False:
                        new_vouchers_list.append(p)
                except:
                    pass
            
            self.published_voucher_configuration['vouchers']    = new_vouchers_list
            self.published_voucher_configuration['count']       = len(new_vouchers_list)
            
        else:
            self.published_voucher_configuration = {'vouchers':[], 'count':0}
        
        self.put()    
    
    def update_published_program(self, program_configuration):
        if is_empty(self.published_program_configuration):
            self.published_program_configuration = {
                                                'programs'  :[program_configuration],
                                                'count'     : 1,
                                                } 
                                            
        else:
            self.flush_dirty_program_configuration()
            existing_programs_list  = self.published_program_configuration.get('programs')
            
            program_key = program_configuration.get('program_key')
            index = 0
            for p in existing_programs_list:
                if p.get('program_key') == program_key:
                    existing_programs_list.pop(index)
                
                index = index+1
            
            existing_programs_list.append(program_configuration)
            
            self.published_program_configuration['programs']    = existing_programs_list
            self.published_program_configuration['count']       = len(existing_programs_list) 
            
        self.put()
        
    def update_published_voucher(self, voucher_configuration):
        if is_empty(self.published_voucher_configuration):
            self.published_voucher_configuration = {
                                                'vouchers'  :[voucher_configuration],
                                                'count'     : 1,
                                                } 
                                            
        else:
            self.flush_dirty_program_configuration()
            existing_vouchers_list  = self.published_voucher_configuration.get('vouchers')
            
            voucher_key = voucher_configuration.get('voucher_key')
            
            index = 0
            for v in existing_vouchers_list:
                if v.get('voucher_key') == voucher_key:
                    existing_vouchers_list.pop(index)
                
                index = index+1
            
            existing_vouchers_list.append(voucher_configuration)
            
            self.published_voucher_configuration['vouchers']    = existing_vouchers_list
            self.published_voucher_configuration['count']       = len(existing_vouchers_list) 
            
        self.put() 
        
    def add_membership(self, membership_configuration):
        if is_empty(self.membership_configuration):
            self.membership_configuration = {
                                                'memberships'  :[membership_configuration],
                                                'count'     : 1,
                                                } 
                                            
        else:
            self.flush_dirty_membership_configuration()
            existing_memberships_list  = self.membership_configuration.get('memberships')
            existing_memberships_list.append(membership_configuration)
            
            self.membership_configuration['memberships']     = existing_memberships_list
            self.membership_configuration['count']           = len(existing_memberships_list) 
            
        self.put()
        
    def update_membership(self, membership_configuration):
        if is_empty(self.membership_configuration):
            self.membership_configuration = {
                                                'memberships'  :[membership_configuration],
                                                'count'     : 1,
                                                } 
                                            
        else:
            self.flush_dirty_membership_configuration()
            existing_memberships_list  = self.membership_configuration.get('memberships')
            
            for idx, em in enumerate(existing_memberships_list):
                if em.get('membership_key') == membership_configuration.get('membership_key'):
                    existing_memberships_list[idx] = membership_configuration
                    break
            
            self.membership_configuration['memberships']     = existing_memberships_list 
            
        self.put()     
        
    def add_tier_membership(self, membership_configuration):
        if is_empty(self.tier_membership_configuration):
            self.tier_membership_configuration = {
                                                'memberships'   :[membership_configuration],
                                                'count'         : 1,
                                                } 
                                            
        else:
            self.flush_dirty_tier_membership_configuration()
            existing_memberships_list  = self.tier_membership_configuration.get('memberships')
            existing_memberships_list.append(membership_configuration)
            
            self.tier_membership_configuration['memberships']     = existing_memberships_list
            self.tier_membership_configuration['count']           = len(existing_memberships_list) 
            
        self.put()
        
    def update_tier_membership(self, membership_configuration):
        if is_empty(self.tier_membership_configuration):
            self.tier_membership_configuration = {
                                                'memberships'   :[membership_configuration],
                                                'count'         : 1,
                                                } 
                                            
        else:
            self.flush_dirty_tier_membership_configuration()
            existing_memberships_list  = self.tier_membership_configuration.get('memberships')
            
            for idx, em in enumerate(existing_memberships_list):
                if em.get('membership_key') == membership_configuration.get('membership_key'):
                    existing_memberships_list[idx] = membership_configuration
                    break
            
            self.tier_membership_configuration['memberships']     = existing_memberships_list
            
        self.put()                
        
    def remove_archieve_program(self, archieve_program_key):
        self.flush_dirty_program_configuration()
        existing_programs_list = self.published_program_configuration['programs']
        
        index = 0
        
        for program in existing_programs_list:
            if program.get('program_key') == archieve_program_key:
                existing_programs_list.pop(index)
            index = index+1
        
        self.published_program_configuration['programs']    = existing_programs_list
        self.published_program_configuration['count']       = len(existing_programs_list)
            
        self.put() 
        
    def remove_archieve_voucher(self, archieve_voucher_key):
        self.flush_dirty_voucher_configuration()
        existing_vouchers_list = self.published_voucher_configuration['vouchers']
        
        index = 0
        
        for voucher in existing_vouchers_list:
            if voucher.get('voucher_key') == archieve_voucher_key:
                existing_vouchers_list.pop(index)
            index = index+1
        
        self.published_voucher_configuration['vouchers']    = existing_vouchers_list
        self.published_voucher_configuration['count']       = len(existing_vouchers_list)
            
        self.put()
        
    def remove_archieve_basic_membership(self, archieve_membership_key):
        self.flush_dirty_membership_configuration()
        existing_memberships_list = self.membership_configuration['memberships']
        
        index = 0
        
        for m in existing_memberships_list:
            if m.get('membership_key') == archieve_membership_key:
                existing_memberships_list.pop(index)
            index = index+1
        
        self.membership_configuration['memberships']    = existing_memberships_list
        self.membership_configuration['count']       = len(existing_memberships_list)
            
        self.put() 
        
    def remove_archieve_tier_membership(self, archieve_membership_key):
        self.flush_dirty_tier_membership_configuration()
        existing_memberships_list = self.tier_membership_configuration['memberships']
        
        index = 0
        
        for m in existing_memberships_list:
            if m.get('membership_key') == archieve_membership_key:
                existing_memberships_list.pop(index)
            index = index+1
        
        self.tier_membership_configuration['memberships']  = existing_memberships_list
        self.tier_membership_configuration['count']             = len(existing_memberships_list)
            
        self.put()                
    
    def update_stat_details(self, stat_dict):
        dashboard_stat_figure = self.dashboard_stat_figure
        logger.debug('update_stat_figure: dashboard_stat_figure=%s', dashboard_stat_figure)
        
        next_updated_datetime = datetime.now() + timedelta(minutes=int(self.stat_figure_update_interval_in_minutes))
        
        logger.debug('update_stat_figure: next_updated_datetime=%s', next_updated_datetime)
        
        dashboard_stat_figure = {
                                'next_updated_datetime' : next_updated_datetime.strftime(self.stat_figure_update_datetime_format),
                                'stat_details'          : stat_dict,
                             }
            
        self.dashboard_stat_figure = dashboard_stat_figure
        self.put()
    
    def get_stat_details(self):
        dashboard_stat_figure = self.dashboard_stat_figure
        
        logger.debug('get_stat_figure: dashboard_stat_figure=%s', dashboard_stat_figure)
        
        if dashboard_stat_figure is not None and dashboard_stat_figure.get('next_updated_datetime'):
            next_updated_datetime = dashboard_stat_figure.get('next_updated_datetime')
            
            logger.debug('get_stat_figure: next_updated_datetime=%s', next_updated_datetime)
            
            if next_updated_datetime:
                next_updated_datetime = datetime.strptime(next_updated_datetime, self.stat_figure_update_datetime_format)
                now = datetime.now()
                if now > next_updated_datetime:
                    return None
                else:
                    return dashboard_stat_figure.get('stat_details')
        
        return None
    
    @staticmethod
    def create(company_name=None, contact_name=None, email=None, mobile_phone=None, office_phone=None, plan_start_date=None, plan_end_date=None, 
               account_code=None, currency_code=None):
        
        if account_code is None:
            account_code    = "%s-%s-%s-%s" % (random_number(4),random_number(4),random_number(4),random_number(4))
            
        merchant_acct   = MerchantAcct(
                                       company_name     = company_name, 
                                       contact_name     = contact_name,
                                       email            = email,
                                       mobile_phone     = mobile_phone,
                                       office_phone     = office_phone,
                                       plan_start_date  = plan_start_date, 
                                       plan_end_date    = plan_end_date,
                                       currency_code    = currency_code
                                       )
        
        logging.debug('account_code=%s', account_code)
        
        merchant_acct.account_code = account_code
        
        merchant_acct.put()
        
        return merchant_acct
    
    @staticmethod
    def get_by_account_code(account_code):
        return MerchantAcct.query(ndb.AND(MerchantAcct.account_code==account_code)).get()
        
    
    @staticmethod
    def list(offset=0, limit=10):
        return MerchantAcct.query().order(-MerchantAcct.registered_datetime).fetch(offset=offset, limit=limit)
    
    def delete_and_related(self):
        
        @ndb.transactional()
        def start_transaction(merchant_acct):
            merchant_user_key_list = MerchantUser.list_by_merchant_account(merchant_acct, keys_only=True)
            if merchant_user_key_list:
                ndb.delete_multi(merchant_user_key_list)
            
            merchant_acct.delete()
            logger.debug('after deleted merchant acct and merchant user')
            
        
        start_transaction(self)
        
class MerchantSentEmail(SentEmail):
    '''
    Merchant account as Ancestor
    '''
    pass

class MerchantUser(UserMin):
    
    '''
    parent is MerchantAcct
    '''
    username                = ndb.StringProperty(required=True)
    permission              = ndb.JsonProperty()
    is_admin                = ndb.BooleanProperty(required=True, default=False)
    
    dict_properties         = ['key', 'user_id', 'name', 'username', 'permission', 
                                'created_datetime', 'active', 'is_admin', 'permission', 
                                'is_super_user', 'is_admin_user', 'is_merchant_user']
    
    @property
    def is_super_user(self):
        return False
    
    @property
    def is_admin_user(self):
        return self.is_admin
    
    @property
    def is_merchant_user(self):
        return True
    
    @property
    def merchant_acct(self):
        return MerchantAcct.fetch(self.key.parent().urlsafe())
    
    @property
    def granted_outlet(self):
        return self.permission.get('granted_outlet')
    
    @property
    def granted_access(self):
        return self.permission.get('granted_access')
    
    @staticmethod
    def update_permission(merchant_user, access_permission, outlet_permission, is_admin=False):
        if access_permission is None:
            access_permission = []
            
        if outlet_permission is None:
            outlet_permission = []
            
        merchant_user.is_admin = is_admin
        merchant_user.permission = {'granted_access': access_permission, 'granted_outlet': outlet_permission}
        merchant_user.put()
    
    @staticmethod
    def create(merchant_acct=None, name=None, 
               username=None,
               password=None):
        
        check_unique_merchant_user = MerchantUser.get_by_username(username)
        
        if check_unique_merchant_user is None:
            user_id = generate_user_id()
            created_user = MerchantUser(
                                parent = merchant_acct.create_ndb_key(),
                                user_id=user_id, 
                                name=name, 
                                username=username
                                )
            
            hashed_password = hash_password(user_id, password)
            created_user.password = hashed_password
                
            created_user.put()
            
            return created_user
        else:
            raise Exception('Username have been used')
    
    @staticmethod
    def count_by_merchant_account(merchant_acct):
        condition_query = MerchantUser.query(ancestor = merchant_acct.create_ndb_key())
        return MerchantUser.count_with_condition_query(condition_query, limit=model_conf.MAX_FETCH_RECORD)  
    
    @staticmethod
    def list_by_merchant_account(merchant_acct, keys_only=False):
        return MerchantUser.query(ancestor = merchant_acct.create_ndb_key()).order(-MerchantUser.created_datetime).fetch(limit=model_conf.MAX_FETCH_RECORD, keys_only=keys_only)
    
    @staticmethod
    def list_all_by_merchant_account(merchant_acct, offset=None, start_cursor=None, return_with_cursor=False, keys_only=False, limit = model_conf.MAX_FETCH_RECORD):
        condition_query =  MerchantUser.query(ancestor = merchant_acct.create_ndb_key()).order(-MerchantUser.created_datetime)
        return MerchantUser.list_all_with_condition_query(
                                        condition_query, 
                                        offset=offset, 
                                        start_cursor=start_cursor, 
                                        return_with_cursor=return_with_cursor, 
                                        keys_only=keys_only, 
                                        limit=limit)
    
    @staticmethod
    def get_by_username(username):
        return MerchantUser.query(ndb.AND(MerchantUser.username==username)).get()
    
    
    @staticmethod
    def get_merchant_acct_by_merchant_user(merchant_user):
        return MerchantAcct.fetch(merchant_user.key.parent().urlsafe())
    
class Outlet(BaseNModel, DictModel, FullTextSearchable):
    merchant_acct           = ndb.KeyProperty(name="merchant_acct", kind=MerchantAcct)
    name                    = ndb.StringProperty(required=True)
    address                 = ndb.StringProperty(required=False)
    office_phone            = ndb.StringProperty(required=False)
    fax_phone               = ndb.StringProperty(required=False)
    email                   = ndb.StringProperty(required=False)
    business_hour           = ndb.StringProperty(required=False)
    is_physical_store       = ndb.BooleanProperty(required=False, default=True)
    geo_location            = ndb.GeoPtProperty(required=False)
    created_datetime        = ndb.DateTimeProperty(required=True, auto_now_add=True)
    
    searchable_field_name   = 'name'
    
    dict_properties         = ['key', 'name', 'address', 'office_phone', 
                                'fax_phone', 'email', 'business_hour', 'is_physical_store', 
                                'geo_location', 'created_datetime']
    
    @property
    def outlet_key(self):
        return self.key.urlsafe()
    
    @staticmethod
    def create(merchant_acct=None,name=None, address=None, email=None, fax_phone=None, 
               office_phone=None, business_hour=None, geo_location=None, is_physical_store=True):
        
        outlet   = Outlet(
                            parent              = merchant_acct.create_ndb_key(),
                            name                = name, 
                            address             = address,
                            email               = email,
                            fax_phone           = fax_phone,
                            office_phone        = office_phone,
                            business_hour       = business_hour,
                            geo_location        = geo_location,
                            is_physical_store   = is_physical_store,
                            )
        
        outlet.put()
        
        return outlet
    
    @staticmethod
    def list_by_merchant_acct(merchant_acct):
        return Outlet.query(ancestor = merchant_acct.create_ndb_key()).fetch(limit=model_conf.MAX_FETCH_RECORD)
    
class MerchantTagging(Tagging):    
    
    @staticmethod
    def create(merchant_acct, label=None, desc=None):
        return MerchantTagging.create_tag(parent=merchant_acct.create_ndb_key(), label=label, desc=desc)
    
    def update(self, label=None, desc=None):
        self.label  = label
        self.desc   = desc
        self.put()
        
    @staticmethod
    def list_by_merchant_account(merchant_acct):
        return MerchantTagging.query(ancestor = merchant_acct.create_ndb_key()).fetch(limit = conf.MAX_FETCH_RECORD)
    
    @staticmethod
    def get_by_merchant_label(merchant_acct, label):
        return MerchantTagging.get_by_label(merchant_acct.create_ndb_key(), label)
    
    


'''
Created on 8 May 2020

@author: jacklok
'''
from google.cloud import ndb
from trexmodel.models.datastore.user_models import UserMin
from trexlib.utils.security_util import generate_user_id, hash_password


class SuperUser(UserMin):
    email                           = ndb.StringProperty(required=False)
    created_datetime                = ndb.DateTimeProperty(required=False, auto_now_add=True)
    
    dict_properties         = ['user_id', 'name', 'email', 'gravatar_url', 'active', 'locked', 
                               'is_super_user', 'is_admin_user', 'is_merchant_user', 'created_datetime']
    
    @property
    def is_super_user(self):
        return True
    
    @property
    def is_admin_user(self):
        return False
    
    @property
    def is_merchant_user(self):
        return False
    
    @classmethod
    def new_super_user_id(cls):
        return 'superuser'
    
    @classmethod
    def create(cls, name=None, email=None, password=None, active=False):
        
        user_id = cls.new_super_user_id()
        created_user = cls(user_id=user_id, name=name, email=email, active=active)
        
        hashed_password         = hash_password(user_id, password)
        created_user.password   = hashed_password
            
        created_user.put()
        return created_user
    
    @classmethod
    def list(cls, offset=0, limit=10):
        return cls.query().order(-cls.joined_date).fetch(offset=offset, limit=limit)
    
    @classmethod
    def get_by_email(cls, email):
        return cls.query(ndb.AND(cls.email==email)).get()
    
class AdminUser(SuperUser):
    permission              = ndb.JsonProperty()
    is_superuser            = ndb.BooleanProperty(required=True, default=False)
    dict_properties         = ['user_id', 'name', 'email', 'gravatar_url', 'active', 'locked', 
                               'permission', 'is_superuser', 'is_super_user', 'is_admin_user', 'is_merchant_user', 'created_datetime']
    
    @classmethod
    def new_super_user_id(cls):
        return generate_user_id()
    
    @property
    def is_super_user(self):
        return self.is_superuser
    
    @property
    def is_admin_user(self):
        return True
    
    @property
    def is_merchant_user(self):
        return False
    
    @staticmethod
    def update_permission(admin_user, permission, is_superuser=False):
        admin_user.is_superuser = is_superuser
        admin_user.permission = {'granted_access': permission}
        admin_user.put()




'''
Created on 29 Apr 2021

@author: jacklok
'''
from trexlib.utils.string_util import is_empty, is_not_empty
from datetime import datetime


def update_reward_summary_with_new_reward(existing_reward_summary, reward_details):
    reward_summary              = existing_reward_summary
    reward_format               = reward_details.get('reward_format')
    reward_amount               = reward_details.get('reward_amount')
    used_reward_amount          = reward_details.get('used_reward_amount')
    latest_expiry_date          = datetime.strptime(reward_details.get('expiry_date'), '%d-%m-%Y').date()
    
    
    existing_latest_expiry_date = None
    new_latest_expiry_date      = latest_expiry_date
    
    if is_empty(reward_summary):
        reward_balance = reward_amount - used_reward_amount
        reward_summary = {
                            reward_format:{
                                            'latest_expiry_date'    : new_latest_expiry_date.strftime('%d-%m-%Y'),
                                            'amount'                : reward_amount
                                            }
                                        
                            }
    else:
        reward_summary_by_reward_format = reward_summary.get(reward_format)
        
        if reward_summary_by_reward_format is None:
            reward_summary_by_reward_format = {
                                                'amount': 0,
                                                }
            new_latest_expiry_date = latest_expiry_date
        else:
        
            existing_latest_expiry_date = reward_summary_by_reward_format.get('latest_expiry_date')
            existing_latest_expiry_date = latest_expiry_date
            
            if latest_expiry_date > existing_latest_expiry_date:  
                new_latest_expiry_date = existing_latest_expiry_date
                
                
        
        reward_summary_by_reward_format['latest_expiry_date']   = new_latest_expiry_date.strftime('%d-%m-%Y')
        
        reward_balance = reward_amount - used_reward_amount
        
        reward_summary_by_reward_format['amount']               = reward_summary_by_reward_format['amount'] + reward_balance
        
        reward_summary[reward_format] = reward_summary_by_reward_format
        
    return reward_summary

def update_reward_summary_with_reverted_reward(existing_reward_summary, reward_details):
    reward_summary              = existing_reward_summary
    reward_format               = reward_details.get('reward_format')
    reward_amount               = reward_details.get('reward_amount')
    used_reward_amount          = reward_details.get('used_reward_amount')
    latest_expiry_date          = datetime.strptime(reward_details.get('expiry_date'), '%d-%m-%Y').date()
    last_transacted_date        = datetime.strptime(reward_details.get('rewarded_date'), '%d-%m-%Y').date()
    
    existing_latest_expiry_date = None
    
    if is_not_empty(reward_summary):
        reward_summary_by_reward_format = reward_summary.get(reward_format)
        
        existing_latest_expiry_date = reward_summary_by_reward_format.get('latest_expiry_date')
        existing_latest_expiry_date = datetime.strptime(existing_latest_expiry_date, '%d-%m-%Y').date()
        
        reward_summary_by_reward_format['amount']  = reward_summary_by_reward_format['amount'] - reward_amount
        
        reward_summary[reward_format] = reward_summary_by_reward_format
        
    return reward_summary



'''
Created on 25 Jan 2021

@author: jacklok
'''
from trexmodel import conf as model_conf
from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel

class UpstreamData(BaseNModel, DictModel):
    table_template_name     = ndb.StringProperty(required=True)
    dataset_name            = ndb.StringProperty(required=True)
    table_name              = ndb.StringProperty(required=True)
    stream_content          = ndb.JsonProperty(required=True)
    is_sent                 = ndb.BooleanProperty(required=True, default=False)
    created_datetime        = ndb.DateTimeProperty(required=True, auto_now_add=True)
    
    
    @staticmethod
    def list_not_send(offset=0, limit=model_conf.MAX_FETCH_RECORD, start_cursor=None,return_with_cursor=False, keys_only=False):
        query = UpstreamData.query(ndb.AND(UpstreamData.is_sent==False))
        
        return UpstreamData.list_all_with_condition_query(query, offset=offset, start_cursor=start_cursor, return_with_cursor=return_with_cursor, keys_only=keys_only, limit=limit)
        
    
    @staticmethod
    def count_not_sent(limit=model_conf.MAX_FETCH_RECORD):
        return UpstreamData.query(ndb.AND(UpstreamData.is_sent==False)).count(limit = limit)
    
    @staticmethod
    def create(dataset_name, table_name, table_template_name, stream_content):
        UpstreamData(
                    table_template_name = table_template_name,
                    dataset_name        = dataset_name,
                    table_name          = table_name,
                    stream_content      = stream_content,
                    ).put()
                    

    
    
                    
    



'''
Created on 16 May 2021

@author: jacklok
'''

from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel, FullTextSearchable
from trexmodel.models.datastore.system_models import SentEmail
from trexmodel.models.datastore.user_models import UserMin
import trexmodel.conf as model_conf
from trexlib.utils.security_util import generate_user_id, hash_password
from trexlib.utils.string_util import random_number
import logging
from datetime import datetime, timedelta
from trexlib.utils.common.date_util import parse_datetime
from trexmodel import conf
from google.auth._default import default
from trexmodel.models.datastore.system_models import Tagging


logger = logging.getLogger('model')

class TestModelBase(BaseNModel, DictModel):
    id                      = ndb.IntegerProperty(required=True, default=1)
    value                   = ndb.StringProperty(required=True)
    modified_datetime       = ndb.DateTimeProperty(required=True, auto_now=True)
    
    @classmethod
    def create(cls, id=1, value='1'):
        test_model = cls(id=id, value=value)
        test_model.put()

    @classmethod
    def get_by_id(cls, id):
        return cls.query(cls.id==id).get()

class TestModelA(TestModelBase):
    pass

        
class TestModelB(TestModelBase):
    pass      


import os, config_path


INTERNAL_MAX_FETCH_RECORD                       = 9999
MAX_FETCH_RECORD_FULL_TEXT_SEARCH               = 1000
MAX_FETCH_RECORD_FULL_TEXT_SEARCH_PER_PAGE      = 10
MAX_FETCH_RECORD                                = 99999999
MAX_FETCH_IMAGE_RECORD                          = 100
MAX_CHAR_RANDOM_UUID4                           = 20
PAGINATION_SIZE                                 = 2
VISIBLE_PAGE_COUNT                              = 10

DEFAULT_GMT_HOURS                               = 8

GENDER_MALE_CODE                                = 'm'
GENDER_FEMALE_CODE                              = 'f'

APPLICATION_ACCOUNT_PROVIDER                    = 'app'

MODEL_PROJECT_ID                                = os.environ['GCLOUD_PROJECT_ID']

DATASTORE_SERVICE_ACCOUNT_KEY_FILEPATH          = os.environ['SERVICE_ACCOUNT_KEY']

DATASTORE_CREDENTIAL_PATH                       = os.path.abspath(os.path.dirname(config_path.__file__)) + '/' + DATASTORE_SERVICE_ACCOUNT_KEY_FILEPATH

MERCHANT_STAT_FIGURE_UPDATE_INTERVAL_IN_MINUTES = os.environ['MERCHANT_STAT_FIGURE_UPDATE_INTERVAL_IN_MINUTES']

HANDLE_DATETIME_WITH_GMT                        = os.environ.get('HANDLE_DATETIME_WITH_GMT')
SERVER_DATETIME_GMT                             = os.environ.get('SERVER_DATETIME_GMT')


'''
Created on 5 Jan 2021

@author: jacklok
'''
from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel, FullTextSearchable
from trexmodel.models.datastore.user_models import User
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet
import trexmodel.conf as model_conf
from trexlib.utils.string_util import is_not_empty, is_empty
import logging
from trexlib.utils.common.cache_util import cache
from trexlib.utils.common.date_util import convert_date_to_datetime
from trexmodel import conf
from six import string_types
from datetime import datetime, timedelta
from trexmodel.models.datastore.customer_model_helpers import update_reward_summary_with_new_reward

logger = logging.getLogger('model')


class Customer(BaseNModel, DictModel, FullTextSearchable):
    '''
    parent is User
    '''
    
    merchant_acct               = ndb.KeyProperty(name="merchant_acct", kind=MerchantAcct)
    outlet                      = ndb.KeyProperty(name="outlet", kind=Outlet)
    merchant_reference_code     = ndb.StringProperty(name="merchant_reference_code", required=False)
    registered_datetime         = ndb.DateTimeProperty(required=False, auto_now_add=True)
    modified_datetime           = ndb.DateTimeProperty(required=False, auto_now=True)
    
    #---------------------------------------------------------------------------
    # User denormalize fields
    #---------------------------------------------------------------------------
    name                        = ndb.StringProperty(required=False)
    mobile_phone                = ndb.StringProperty(required=False)
    email                       = ndb.StringProperty(required=False)
    
    birth_date                  = ndb.DateProperty(required=False, indexed=False)
    birth_date_date_str         = ndb.StringProperty(required=False) 
    gender                      = ndb.StringProperty(required=False)
    reference_code              = ndb.StringProperty(required=True)
    
    mobile_app_installed        = ndb.BooleanProperty(required=False, default=False)
    
    tags_list                   = ndb.StringProperty(repeated=True, write_empty_list=True)
    
    memberships_list            = ndb.StringProperty(repeated=True, write_empty_list=True)
    
    last_transact_date          = ndb.DateProperty(required=False)
    
    tier_membership             = ndb.StringProperty(required=False)
    
    reward_summary              = ndb.JsonProperty()
    
    fulltextsearch_field_name   = 'name'
    
    dict_properties     = ['name', 'mobile_phone', 'email', 'gender', 'birth_date', 'reference_code', 'merchant_reference_code',  
                           'tags_list', 'memberships_list', 'tier_membership', 'registered_merchant_acct', 'reward_summary',
                           'registered_outlet_key', 'registered_merchant_acct_key', 'registered_datetime', 'modified_datetime']
    
    @property
    def registered_user_acct_key(self):
        return self.key.parent().urlsafe()
    
    @property
    def registered_user_acct(self):
        return User.fetch(self.key.parent().urlsafe())
    
    @property
    def registered_merchant_acct(self):
        return MerchantAcct.fetch(self.merchant_acct.urlsafe())
    
    @property
    def registered_merchant_acct_key(self):
        if self.merchant_acct:
            return self.merchant_acct.urlsafe().decode("utf-8")
    
    @property
    def registered_outlet(self):
        if self.outlet:
            return Outlet.fetch(self.outlet.urlsafe())
    
    @property
    def registered_outlet_key(self):
        if self.outlet:
            return self.outlet.urlsafe().decode("utf-8")
    
    @staticmethod    
    def update_KPI(customer_acct, tags_list=None, memberships_list=None, tier_membership=None):
        if isinstance(tags_list, string_types):
            if is_not_empty(tags_list):
                tags_list = tags_list.split(',')
            else:
                tags_list = []
                
        if isinstance(memberships_list, string_types):
            if is_not_empty(memberships_list):
                memberships_list = memberships_list.split(',')
            else:
                memberships_list = []
                
                
        
        customer_acct.tags_list = tags_list
        customer_acct.memberships_list = memberships_list
        customer_acct.tier_membership = tier_membership
        customer_acct.put()
        
        
    @classmethod
    @cache.memoize(timeout=60)
    def get_by_merchant_reference_code(cls, merchant_reference_code):
        return cls.query(cls.merchant_reference_code==merchant_reference_code).get()
    
    @classmethod
    def create(cls, merchant_acct=None, outlet=None, name=None, email=None, mobile_phone=None, merchant_reference_code=None,
               password=None):
        
        created_user = User.create(name=name, email=email, mobile_phone=mobile_phone, 
                           password=password)
        
        created_user.put()
        
        created_customer = cls.create_from_user(merchant_acct, outlet, created_user)
        
        return created_customer
    
    @classmethod
    def update(cls, customer=None, outlet=None, **kwargs):
        
        customer.outlet = outlet.create_ndb_key()
        
        logger.debug('**kwargs=%s', kwargs)
        
        for key, value in kwargs.items():
            setattr(customer, key, value)
        
        user_acct = customer.registered_user_acct
        
        user_acct.name                   = customer.name
        user_acct.email                  = customer.email
        user_acct.mobile_phone           = customer.mobile_phone
        user_acct.birth_date             = customer.birth_date
        user_acct.birth_date_date_str    = customer.birth_date_date_str
        user_acct.gender                 = customer.gender
        
        user_acct.put()
        customer.put()
        
    @classmethod
    def create_from_user(cls, merchant_acct, outlet, user_acct, merchant_reference_code=None):
        birth_date = user_acct.birth_date
        birth_date_date_str = user_acct.birth_date_date_str
        
        created_customer = cls(parent=user_acct.create_ndb_key(), 
                               outlet                   = outlet.create_ndb_key(), 
                               name                     = user_acct.name, 
                               email                    = user_acct.email, 
                               mobile_phone             = user_acct.mobile_phone, 
                               gender                   = user_acct.gender, 
                               reference_code           = user_acct.reference_code, 
                               birth_date               = birth_date, 
                               birth_date_date_str      = birth_date_date_str, 
                               merchant_reference_code  = merchant_reference_code,
                               merchant_acct            = merchant_acct.create_ndb_key()
                           )
        
        created_customer.put()
        
        return created_customer
    
    @classmethod
    def list_merchant_customer(cls, merchant_acct, offset=0, limit=conf.PAGINATION_SIZE, start_cursor=None, return_with_cursor=False):
        query = cls.query(ndb.AND(cls.merchant_acct==merchant_acct.create_ndb_key()))
        
        return cls.list_all_with_condition_query(query, offset=offset, limit=limit, start_cursor=start_cursor, return_with_cursor=return_with_cursor)
    
    @classmethod
    def list_by_user_account(cls, user_acct):
        return cls.query(ancestor=user_acct.create_ndb_key()).fetch(limit=conf.MAX_FETCH_RECORD)
    
    @classmethod
    def count_merchant_customer(cls, merchant_acct):
        if merchant_acct:
            query = cls.query(ndb.AND(cls.merchant_acct==merchant_acct.create_ndb_key()))
        else:
            query = cls.query()
        
        return cls.count_with_condition_query(query)
    
    @classmethod
    def search_merchant_customer(cls, merchant_acct, name=None, email=None, mobile_phone=None, 
                                 reference_code=None, merchant_reference_code=None, merchant_tagging=None,
                                 registered_date_start=None, registered_date_end=None,
                                 offset=0, start_cursor=None, limit=model_conf.MAX_FETCH_RECORD):
        
        search_text_list = None
        query = cls.query(ndb.AND(cls.merchant_acct==merchant_acct.create_ndb_key()))
        
        if is_not_empty(email):
            query = query.filter(cls.email==email)
            
        elif is_not_empty(mobile_phone):
            query = query.filter(cls.mobile_phone==mobile_phone)
            
        elif is_not_empty(reference_code):
            query = query.filter(cls.reference_code==reference_code)
            
        elif is_not_empty(merchant_reference_code):
            query = query.filter(cls.merchant_reference_code==merchant_reference_code)
                    
        elif is_not_empty(merchant_tagging):
            query = query.filter(cls.tags_list==merchant_tagging)
        
        elif is_not_empty(name):
            search_text_list = name.split(' ')
        
        elif is_not_empty(registered_date_start) or is_not_empty(registered_date_end):
            
            if is_not_empty(registered_date_start):
                registered_datetime_start = convert_date_to_datetime(registered_date_start)
                
                query = query.filter(cls.registered_datetime>=registered_datetime_start)
            
            if is_not_empty(registered_date_end):
                registered_datetime_end = convert_date_to_datetime(registered_date_end)
            
                query = query.filter(cls.registered_datetime<registered_datetime_end)
        
        total_count                         = cls.full_text_count(search_text_list, query, conf.MAX_FETCH_RECORD_FULL_TEXT_SEARCH)
        
        (search_results, next_cursor)       = cls.full_text_search(search_text_list, query, offset=offset, 
                                                                   start_cursor=start_cursor, return_with_cursor=True, 
                                                                   limit=limit)
        
        return (search_results, total_count, next_cursor)
    
    def update_from_user_acct(self, user_acct):
        
        self.name                   = user_acct.name
        self.email                  = user_acct.email
        self.mobile_phone           = user_acct.mobile_phone
        self.birth_date             = user_acct.birth_date
        self.birth_date_date_str    = user_acct.birth_date_date_str
        self.gender                 = user_acct.gender
        self.put()
    
    @staticmethod
    def count_by_last_transact_date(merchant_acct, last_transact_in_day=7):
        checking_date = datetime.now().date() - timedelta(days=last_transact_in_day)
        
        logger.debug('count_by_last_transact_date: checking_date=%s', checking_date)
        
        query = Customer.query(ndb.AND(
                                                    Customer.merchant_acct==merchant_acct.create_ndb_key(),
                                                    Customer.last_transact_date >= checking_date   
                                        ))
        
        return Customer.count_with_condition_query(query)
    
    


'''
Created on 19 Feb 2021

@author: jacklok
'''
from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel
import trexmodel.conf as model_conf
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet,\
    MerchantUser
from trexmodel.models.datastore.coporate_models import CorporateAcct
import logging
from trexlib.utils.common.cache_util import cache
from trexmodel import conf, program_conf
from datetime import datetime


logger = logging.getLogger('model')


class ProgramBase(BaseNModel, DictModel):
    reward_base             = ndb.StringProperty(required=True, choices=set(program_conf.REWARD_BASE_SET))
    reward_format           = ndb.StringProperty(required=False, choices=set(program_conf.REWARD_FORMAT_SET))
    completed_status        = ndb.StringProperty(required=True, choices=set(program_conf.PROGRAM_STATUS))
    giveaway_method         = ndb.StringProperty(required=False, default=program_conf.PROGRAM_REWARD_GIVEAWAY_METHOD_AUTO)
    start_date              = ndb.DateProperty(required=True)
    end_date                = ndb.DateProperty(required=True)
    desc                    = ndb.StringProperty(required=False)
    program_settings        = ndb.JsonProperty(required=True)
    created_datetime        = ndb.DateTimeProperty(required=True, auto_now_add=True)
    modified_datetime       = ndb.DateTimeProperty(required=True, auto_now=True)
    published_datetime      = ndb.DateTimeProperty(required=False)
    archived_datetime       = ndb.DateTimeProperty(required=False)
    
    
    created_by              = ndb.KeyProperty(name="created_by", kind=MerchantUser)
    created_by_username     = ndb.StringProperty(required=False)
    modified_by             = ndb.KeyProperty(name="modified_by", kind=MerchantUser)
    modified_by_username    = ndb.StringProperty(required=False)
    
    enabled                 = ndb.BooleanProperty(default=True)
    archived                = ndb.BooleanProperty(default=False)
    
    dict_properties         = ['reward_base', 'giveaway_method', 'reward_format', 'completed_status', 'start_date', 'end_date', 'desc', 'program_settings', 
                               'created_datetime', 'modified_datetime',  'enabled','completed_status','is_enabled', 'is_disabled', 'is_review_state', 
                               'is_published', 'archived', 'is_reward_amount_required', 'completed_prograss_in_percentage', 'completed_status_index',
                               'exclusive_tags_list', 'exclusive_memberships_list', 'exclusive_tier_memberships_list',
                               'is_recurring_scheme', 'is_multitier_scheme', 'is_multitier_membership',
                               'expiration_type', 'expiration_date', 'expiration_value', 'is_expiration_date_type',
                               'created_by_username', 'modified_by_username']
    
    @property
    def is_enabled(self):
        return self.enabled
    
    @property
    def is_disabled(self):
        return self.enabled==False
    
    @property
    def is_archived(self):
        return self.archived
    
    @property
    def is_published(self):
        return self.completed_status == program_conf.PROGRAM_STATUS_PUBLISH
    
    @property
    def is_review_state(self):
        return program_conf.is_existing_program_status_final_state(self.completed_status)
    
    @property
    def completed_prograss_in_percentage(self):
        return program_conf.program_completed_progress_percentage(self.completed_status)
    
    @property
    def completed_status_index(self):
        return program_conf.get_program_completed_status_index(self.completed_status)
    
    @property
    def spending_currency(self):
        return self.program_settings.get('scheme').get('spending_currency')
    
    @property
    def reward_amount(self):
        return self.program_settings.get('scheme').get('reward_amount')
    
    @property
    def is_reward_amount_required(self):
        return self.reward_base!=program_conf.REWARD_FORMAT_VOUCHER
    
    @property
    def is_multitier_scheme(self):
        if self.program_settings and self.program_settings.get('scheme'):
            return self.program_settings.get('scheme').get('is_multitier_scheme') or False
    
    @property
    def is_multitier_membership(self):
        return self.program_settings and self.program_settings.get('is_multitier_membership')
        
    @property
    def is_recurring_scheme(self):
        if self.program_settings and self.program_settings.get('scheme'):
            return self.program_settings.get('scheme').get('is_recurring_scheme') or True
        
    @property
    def expiration_type(self):
        if self.program_settings and self.program_settings.get('scheme'):
            return self.program_settings.get('scheme').get('expiration_type')
        
    @property
    def is_expiration_date_type(self):
        if self.program_settings and self.program_settings.get('scheme'):
            return self.program_settings.get('scheme').get('expiration_type') == program_conf.REWARD_EXPIRATION_TYPE_SPECIFIC_DATE    
        
    @property
    def expiration_date(self):
        if self.program_settings and self.program_settings.get('scheme'):
            return self.program_settings.get('scheme').get('expiration_date')
        
    @property
    def expiration_value(self):
        if self.program_settings and self.program_settings.get('scheme'):
            return self.program_settings.get('scheme').get('expiration_value')                
    
    
    @property
    def exclusive_tags_list(self):
        if self.program_settings.get('exclusivity') and self.program_settings.get('exclusivity').get('tags'):
            return ','.join(self.program_settings.get('exclusivity').get('tags')) or ''
        
    @property
    def exclusive_memberships_list(self):
        if self.program_settings.get('exclusivity') and self.program_settings.get('exclusivity').get('memberships'):
            return ','.join(self.program_settings.get('exclusivity').get('memberships')) or ''
        
    @property
    def exclusive_tier_memberships_list(self):
        if self.program_settings.get('exclusivity') and self.program_settings.get('exclusivity').get('tier_memberships'):
            return ','.join(self.program_settings.get('exclusivity').get('tier_memberships')) or ''        
    
    def to_program_configuration(self):
        program_configuration = {
                                'program_key'       : self.key_in_str,
                                'desc'              : self.desc,
                                'reward_base'       : self.reward_base,
                                'reward_format'     : self.reward_format,
                                'start_date'        : self.start_date.strftime('%d-%m-%Y'),
                                'end_date'          : self.end_date.strftime('%d-%m-%Y'),    
                                'program_settings'  : self.program_settings,
                                'is_published'      : self.is_published,   
                                }
        
        return program_configuration
    
    def to_program_setttings(self):
        program_setttings = {
                                'program_key'       : self.key_in_str,
                                'desc'              : self.desc,
                                'reward_base'       : self.reward_base,
                                'reward_format'     : self.reward_format,
                                'start_date'        : self.start_date.strftime('%d-%m-%Y'),
                                'end_date'          : self.end_date.strftime('%d-%m-%Y'),    
                                'program_settings'  : self.program_settings,
                                'is_published'      : self.is_published,   
                                }
        
        return program_setttings
    
    @classmethod
    def update_program_base_data(cls, program, reward_base=None, reward_format=None, giveaway_method=None, 
                                 desc=None,start_date=None, end_date=None, modified_by=None):
        
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        program.reward_base            = reward_base
        program.reward_format          = reward_format
        program.giveaway_method        = giveaway_method
        program.desc                   = desc
        program.start_date             = start_date
        program.end_date               = end_date
        program.modified_by            = modified_by.create_ndb_key()
        program.modified_by_username   = modified_by_username
        program.completed_status       = program_conf.PROGRAM_STATUS_REWARD_BASE
        
        
        program.put()
        
        return program
    '''
    @classmethod
    def update_prorgram_reward_format_data(cls, program, reward_format=None, modified_by=None):
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        program.reward_format          = reward_format
        program.modified_by            = modified_by.create_ndb_key()
        program.modified_by_username   = modified_by_username
        program.completed_status       = program_conf.PROGRAM_STATUS_REWARD_FORMAT
        
        
        program.put()
        
        return program
    '''
    @classmethod
    def update_prorgram_reward_scheme_data(cls, program, reward_scheme_configuration={}, modified_by=None):
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        if program.program_settings is None:
            program_settings = {
                            'scheme'                : [reward_scheme_configuration]
                            }
            
        else:
            program_settings = program.program_settings
            program_settings['scheme']              = reward_scheme_configuration
                
        program.modified_by            = modified_by.create_ndb_key()
        program.modified_by_username   = modified_by_username
        program.completed_status       = program_conf.PROGRAM_STATUS_REWARD_SCHEME
        program.program_settings       = program_settings
        
        
        program.put()
        
        return program
    
    
    @classmethod
    def add_program_voucher(cls, program, voucher_configuration, modified_by=None):
        
        modified_by_username = None
        
        voucher_configuration_list = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        if program.program_settings is None:
            voucher_configuration_list  = []
            program_settings            = {}
            
        else:
            program_settings = program.program_settings
            voucher_configuration_list = program_settings.get('reward_items') or []
            
        
        voucher_configuration_list.append(voucher_configuration)
        program_settings['reward_items']    = voucher_configuration_list
        
        program.modified_by            = modified_by.create_ndb_key()
        program.modified_by_username   = modified_by_username
        program.program_settings       = program_settings
        
        
        program.put()
        
        return program
    
    @classmethod
    def add_program_tier_membership(cls, program, membership_configuration, modified_by=None):
        
        modified_by_username = None
        
        membership_configuration_list = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        if program.program_settings is None:
            membership_configuration_list  = []
            program_settings            = {}
            
        else:
            program_settings = program.program_settings
            membership_configuration_list = program_settings.get('tier_memberships') or []
            
        
        membership_configuration_list.append(membership_configuration)
        
        program_settings['tier_memberships']    = membership_configuration_list
        
        program.modified_by            = modified_by.create_ndb_key()
        program.modified_by_username   = modified_by_username
        program.program_settings       = program_settings
        
        
        program.put()
        
        return program
    
    @classmethod
    def add_program_membership(cls, program, membership_configuration, modified_by=None):
        
        modified_by_username = None
        
        membership_configuration_list = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        if program.program_settings is None:
            membership_configuration_list   = []
            program_settings                = {}
            
        else:
            program_settings = program.program_settings
            membership_configuration_list = program_settings.get('memberships') or []
            
        
        membership_configuration_list.append(membership_configuration)
        
        program_settings['memberships']    = membership_configuration_list
        
        program.modified_by            = modified_by.create_ndb_key()
        program.modified_by_username   = modified_by_username
        program.program_settings       = program_settings
        
        
        program.put()
        
        return program
    
    @classmethod
    def update_prorgram_membership_type_data(cls, program, is_multitier_membership=False, modified_by=None):
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        program_settings = program.program_settings
        program_settings['is_multitier_membership'] = is_multitier_membership
            
        program.modified_by            = modified_by.create_ndb_key()
        program.modified_by_username   = modified_by_username
        program.program_settings       = program_settings
        
        program.put()
        
        return program
    
    @classmethod
    def update_prorgram_exclusivity_data(cls, program, exclusivity_configuration={}, modified_by=None):
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        program_settings = program.program_settings
        program_settings['exclusivity'] = exclusivity_configuration
            
        program.modified_by            = modified_by.create_ndb_key()
        program.modified_by_username   = modified_by_username
        program.completed_status       = program_conf.PROGRAM_STATUS_REWARD_EXCLUSIVITY
        program.program_settings       = program_settings
        
        program.put()
        
        return program
    
    @property
    def reward_items(self):
        return self.program_settings.get('reward_items')
    
    @property
    def exclusivity_configuration(self):
        return self.program_settings.get('exclusivity')
    
    @classmethod
    def remove_program_voucher(cls, program, voucher_index, modified_by=None):
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        program_settings            = program.program_settings
        program_voucher_listing     = program_settings.get('reward_items')
        
        index = 0
        
        for voucher in program_voucher_listing:
            if voucher.get('voucher_index') == voucher_index:
                program_voucher_listing.pop(index)
                break
            
            index = index +1
    
        program_settings['reward_items']    = program_voucher_listing      
        program.program_settings            = program_settings
        program.modified_by                 = modified_by.create_ndb_key()
        program.modified_by_username        = modified_by_username
        program.put()
        
    @classmethod
    def remove_program_tier_membership(cls, program, membership_key, modified_by=None):
        modified_by_username = None
        
        if is_not_empty(modified_by):
            if isinstance(modified_by, MerchantUser):
                modified_by_username = modified_by.username
        
        program_settings            = program.program_settings
        tier_membership_listing     = program_settings.get('tier_memberships')
        
        index = 0
        
        for membership in tier_membership_listing:
            if membership.get('membership_key') == membership_key:
                tier_membership_listing.pop(index)
                break
            
            index = index +1
    
        program_settings['tier_memberships']    = tier_membership_listing      
        program.program_settings                = program_settings
        program.modified_by                     = modified_by.create_ndb_key()
        program.modified_by_username            = modified_by_username
        program.put()    
        
    @classmethod
    def publish_program(cls, program):
        program.completed_status = program_conf.PROGRAM_STATUS_PUBLISH
        program.published_datetime = datetime.now()
        program.put()
        
        
        merchant_acct = program.merchant_acct
        merchant_acct.update_published_program(program.to_program_configuration())
        
    @classmethod
    def archive_program(cls, program):
        program.archived = True
        program.archived_datetime = datetime.now()
        program.put()
        
        merchant_acct = program.merchant_acct
        merchant_acct.remove_archieve_program(program.key_in_str)    
    
    @classmethod
    def enable(cls, program):
        program.enabled = True
        program.put()
        
    @classmethod
    def disable(cls, program):
        program.enabled = False
        program.put()  
    
class MerchantProgram(ProgramBase):
    #merchant_acct           = ndb.KeyProperty(name="merchant_acct", kind=MerchantAcct) 
    
    
    @staticmethod
    def create(merchant_acct, reward_base=None, giveaway_method=None, reward_format=None, desc=None, start_date=None, end_date=None, created_by=None):
        created_by_username = None
        if is_not_empty(created_by):
            if isinstance(created_by, MerchantUser):
                created_by_username = created_by.username
        
        
        merchant_program =  MerchantProgram(
                                        parent              = merchant_acct.create_ndb_key(),
                                        reward_base         = reward_base,
                                        giveaway_method     = giveaway_method,
                                        reward_format       = reward_format,
                                        desc                = desc,
                                        start_date          = start_date,
                                        end_date            = end_date,
                                        created_by          = created_by.create_ndb_key(),
                                        created_by_username = created_by_username,
                                        completed_status    = program_conf.PROGRAM_STATUS_REWARD_BASE,
                                        program_settings    = {},
                                        )
        
        merchant_program.put()
        return merchant_program
    
    
    
    @property
    def merchant_acct(self):
        return MerchantAcct.fetch(self.key.parent().urlsafe())
    
      
    
    @staticmethod
    def list_by_merchant_account(merchant_acct):
        return MerchantProgram.query(ndb.AND(MerchantProgram.archived!=True), ancestor=merchant_acct.create_ndb_key()).fetch(limit=model_conf.MAX_FETCH_RECORD)
    
    @staticmethod
    def list_archived_by_merchant_account(merchant_acct):
        return MerchantProgram.query(ndb.AND(MerchantProgram.archived==True), ancestor=merchant_acct.create_ndb_key()).fetch(limit=model_conf.MAX_FETCH_RECORD)

class BrandProgram(ProgramBase): 
    #corporate_acct           = ndb.KeyProperty(name="corporate_acct", kind=CorporateAcct)
    pass


           





'''
Created on 22 Apr 2021

@author: jacklok
'''
from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel, FullTextSearchable
from trexmodel.models.datastore.user_models import User
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet, MerchantUser
from trexmodel.models.datastore.program_models import MerchantProgram
from trexmodel.models.datastore.voucher_models import MerchantVoucher
from trexlib.utils.string_util import is_empty
import logging
from trexlib.utils.common.cache_util import cache
from trexmodel import conf, program_conf
from trexlib.utils.string_util import random_string
from datetime import datetime, timedelta
from trexmodel.utils.model.model_util import generate_transaction_id


logger = logging.getLogger('model')


class RewardEntitlement(BaseNModel, DictModel):
    effective_date              = ndb.DateProperty(required=True)
    expiry_date                 = ndb.DateProperty(required=True)
    
    transaction_id              = ndb.StringProperty(required=True)
    invoice_id                  = ndb.StringProperty(required=False)
    
    rewarded_datetime           = ndb.DateTimeProperty(required=True, auto_now_add=True)
    rewarded_by                 = ndb.KeyProperty(name="rewarded_by", kind=MerchantUser)
    rewarded_by_username        = ndb.StringProperty(required=False)
    
    reward_datetime_provided    = ndb.BooleanProperty(required=False, default=False)
    
    status                      = ndb.StringProperty(required=False, default=program_conf.REWARD_STATUS_VALID)
    
    @classmethod
    def list_by_transaction_id(cls, transaction_id):
        return cls.query(cls.transaction_id==transaction_id).fetch(limit=conf.MAX_FETCH_RECORD)
    
class CustomerEntitledReward(RewardEntitlement):
    '''
    Customer as ancestor
    '''
    
    merchant_acct               = ndb.KeyProperty(name="merchant_acct", kind=MerchantAcct)
    user_acct                   = ndb.KeyProperty(name="user_acct", kind=User)
    transact_outlet             = ndb.KeyProperty(name="transact_outlet", kind=Outlet)
    reward_program              = ndb.KeyProperty(name="reward_program", kind=MerchantProgram)
    
    @property
    def rewarded_by_merchant_acct_entity(self):
        return MerchantAcct.fetch(self.merchant_acct.urlsafe())
    
    @property
    def reward_format(self):
        pass
    
    @property
    def reward_format_key(self):
        pass
    
    @property
    def reward_format_label(self):
        pass
    
    @property
    def rewarded_datetime_with_gmt(self):
        gmt_hour = self.rewarded_by_merchant_acct_entity.gmt_hour
        
        return self.rewarded_datetime + timedelta(hours=gmt_hour)
    
    @property
    def is_used(self):
        pass
    
    def revert(self, reverted_by, reverted_datetime=None):
        self.status = program_conf.REWARD_STATUS_REVERTED
        if reverted_datetime is None:
            reverted_datetime = datetime.now()
        
        self.reverted_datetime      = reverted_datetime
        self.reverted_by            = reverted_by.create_ndb_key()
        self.reverted_by_username   = reverted_by.username
        self.put()
    
    @classmethod
    def list_by_customer(cls, customer, status=program_conf.REWARD_STATUS_VALID, limit = conf.MAX_FETCH_RECORD):
        return cls.query(ndb.AND(cls.status==status), ancestor=customer.create_ndb_key()).fetch(limit=limit)

class CustomerCountableReward(CustomerEntitledReward):
    reward_amount               = ndb.IntegerProperty(required=True, default=0)
    used_reward_amount          = ndb.IntegerProperty(required=True, default=0)
    
    @property
    def is_used(self):
        return self.used_reward_amount>0
    
    @property
    def reward_brief(self):
        return 'Entitle {reward_amount} {reward_format}'.format(reward_amount=self.reward_amount, reward_format=self.reward_format_label)
    
    def to_reward_summary(self):
        return {
                'reward_format'         : self.reward_format, 
                'reward_amount'         : self.reward_amount,
                'used_reward_amount'    : self.used_reward_amount,
                'expiry_date'           : self.expiry_date.strftime('%d-%m-%Y'),
                }
    
    @classmethod
    def create(cls, reward_amount, customer_acct, transact_outlet,
               effective_date=None, expiry_date=None, 
               transaction_id=None, invoice_id=None,
               rewarded_by=None):
        
        if is_empty(transaction_id):
            transaction_id = generate_transaction_id()
            
        if effective_date is None:
            effective_date = datetime.today()
        
        customer_reward = cls(
                                            parent                  = customer_acct.create_ndb_key(),
                                            reward_amount           = reward_amount,
                                            transaction_id          = transaction_id,
                                            invoice_id              = invoice_id,
                                            
                                            merchant_acct           = customer_acct.merchant_acct,
                                            transact_outlet         = transact_outlet.create_ndb_key(),
                                            
                                            effective_date          = effective_date,
                                            expiry_date             = expiry_date,
                                            
                                            rewarded_by             = rewarded_by.create_ndb_key(),
                                            rewarded_by_username    = rewarded_by.username,
                                            
                                            )
        
        customer_reward.put()
        
        return customer_reward
    
class RevertedCustomerCountableReward(CustomerCountableReward):
    reverted_datetime           = ndb.DateTimeProperty(required=True, auto_now_add=True)
    reverted_by                 = ndb.KeyProperty(name="reverted_by", kind=MerchantUser)
    reverted_by_username        = ndb.StringProperty(required=False)
    
    @classmethod
    def create(cls, customer_reward, reverted_by):
        reverted_customer_reward = cls(
                            parent                  = customer_reward.key.parent(),
                            reward_amount           = customer_reward.reward_amount,
                            transaction_id          = customer_reward.transaction_id,
                            invoice_id              = customer_reward.invoice_id,
                            
                            merchant_acct           = customer_reward.merchant_acct,
                            transact_outlet         = customer_reward.transact_outlet,
                            
                            effective_date          = customer_reward.effective_date,
                            expiry_date             = customer_reward.expiry_date,
                            
                            rewarded_by             = customer_reward.rewarded_by,
                            rewarded_by_username    = customer_reward.rewarded_by_username,
                            
                            reverted_by             = reverted_by.create_ndb_key(),
                            reverted_by_username    = reverted_by.username,
                            
                            )
        
        reverted_customer_reward.put()
        
class CustomerPointReward(CustomerCountableReward):
    
    @property
    def reward_format(self):
        return program_conf.REWARD_FORMAT_POINT
    
    @property
    def reward_format_label(self):
        return 'point(s)'

class CustomerStampReward(CustomerCountableReward):
    
    @property
    def reward_format(self):
        return program_conf.REWARD_FORMAT_STAMP
    
    @property
    def reward_format_label(self):
        return 'stamp(s)'
    

class RevertedCustomerPointReward(RevertedCustomerCountableReward):
    pass

class RevertedCustomerStampReward(RevertedCustomerCountableReward):
    pass
    
class CustomerEntitledVoucher(CustomerEntitledReward):
    
    entitled_voucher            = ndb.KeyProperty(name='entitled_voucher', kind=MerchantVoucher)
    
    redeem_code                 = ndb.StringProperty(required=False)
    redeemed_datetime           = ndb.DateTimeProperty(required=False)
    
    redeemed_by_outlet          = ndb.KeyProperty(name='redeemed_by_outlet', required=False)
    redeemed_by                 = ndb.KeyProperty(name="redeemed_by", kind=MerchantUser)
    redeemed_by_username        = ndb.StringProperty(required=False)
    
    reverted_datetime           = ndb.DateTimeProperty(required=True, auto_now_add=True)
    reverted_by                 = ndb.KeyProperty(name="reverted_by", kind=MerchantUser)
    reverted_by_username        = ndb.StringProperty(required=False)
    
    redeemed_transaction_id     = ndb.StringProperty(required=False)
    
    removed_datetime            = ndb.DateTimeProperty(required=False)
    removed_by_username         = ndb.StringProperty(required=False)
    
    use_online                  = ndb.BooleanProperty(required=False, default=False)
    use_in_store                = ndb.BooleanProperty(required=False, default=False)
    
    dict_properties     = ['redeem_code', 'voucher_configuration', 'rewarded_by_username', 'rewarded_datetime', 'transaction_id', 
                           'invoice_id', 'merchant_reference_code', 'tags_list', 'status',
                           'registered_outlet_key', 'registered_merchant_acct_key', 'registered_datetime', 'modified_datetime']
    
    def redeem(self, redeemed_by, redeemed_datetime=None):
        self.status = program_conf.REWARD_STATUS_REDEEMED   
        
        if redeemed_datetime is None:
            redeemed_datetime = datetime.now()
        
        self.redeemed_datetime      = redeemed_datetime
        self.redeemed_by            = redeemed_by.create_ndb_key()
        self.redeemed_by_username   = redeemed_by.username
        self.put() 
    
    @property
    def is_used(self):
        return self.status == program_conf.REWARD_STATUS_REDEEMED
    
    @property
    def transact_outlet_summary(self):
        return Outlet.fetch(self.transact_outlet.urlsafe())
    
    @property
    def voucher_configuration(self):
        voucher = MerchantVoucher.fetch(self.entitled_voucher.urlsafe())
        if voucher:
            return voucher.to_configuration()
    
    @property
    def entitled_voucher_key(self):
        return self.entitled_voucher.urlsafe()
    
    @staticmethod
    def create(entitled_voucher, customer_acct, transact_outlet, 
               effective_date=None, expiry_date=None,
               transaction_id=None, invoice_id=None,
               rewarded_by=None, rewarded_datetime=None):
        
        redeem_code = random_string(program_conf.REDEEM_CODE_LENGTH, is_human_mistake_safe=True)
        
        if is_empty(transaction_id):
            transaction_id = generate_transaction_id()
        
        customer_entiteld_voucher = CustomerEntitledVoucher(
                                            parent                  = customer_acct.create_ndb_key(),
                                            entitled_voucher        = entitled_voucher.create_ndb_key(),
                                            
                                            effective_date          = effective_date,
                                            expiry_date             = expiry_date,
                                            
                                            redeem_code             = redeem_code,
                                            transaction_id          = transaction_id,
                                            invoice_id              = invoice_id,
                                            
                                            merchant_acct           = customer_acct.merchant_acct,
                                            transact_outlet         = transact_outlet.create_ndb_key(),
                                            
                                            rewarded_by             = rewarded_by.create_ndb_key(),
                                            rewarded_by_username    = rewarded_by.username,
                                            
                                            rewarded_datetime       = rewarded_datetime,
                                            
                                            )
        
        customer_entiteld_voucher.put()
        
        
    @staticmethod
    def get_by_redeem_code(redeem_code):
        return  CustomerEntitledVoucher.query(CustomerEntitledVoucher.redeem_code==redeem_code).get()
    
    
    
class VoucherRewardDetails(object):    
    
    def __init__(self, voucher_key, reward_amount, expiry_date, rewarded_datetime, merchant_acct=None):
        self.merchant_acct      = merchant_acct
        self.voucher_key        = voucher_key
        self.reward_amount      = reward_amount
        self.expiry_date        = expiry_date
        self.rewarded_datetime  = rewarded_datetime
        
    def __repr__(self, *args, **kwargs):
        return 'voucher_key=%s, reward_amount=%d, expiry_date=%s, rewarded_datetime=%s' % (self.voucher_key, self.reward_amount, self.expiry_date, self.rewarded_datetime)    
        
    @property
    def reward_format(self):
        return program_conf.REWARD_FORMAT_VOUCHER
    
    @property
    def reward_format_key(self):
        return self.voucher_key
    
    @property
    def rewarded_datetime_with_gmt(self):
        gmt_hour = self.merchant_acct.gmt_hour
        
        return self.rewarded_datetime + timedelta(hours=gmt_hour)
    


'''
Created on 19 Feb 2021

@author: jacklok
'''
from trexmodel.models.datastore.program_models import MerchantProgram

class MerchantRewardProgramSpendingBase(MerchantProgram):
    
    @classmethod
    def create(cls, merchant_acct, desc=None, start_date=None, end_date=None, program_settings=None):
        program = cls(
                    merchant_acct   = merchant_acct.create_ndb_key(),
                    desc            = desc,
                    start_date      = start_date,
                    end_date        = end_date,
                    
                    )
        program.put()



'''
Created on 10 Apr 2020

@author: jacklok
'''
from google.cloud import ndb
from trexmodel.models.datastore.ndb_models import BaseNModel, DictModel
#from trexlib.utils.common.cache_util import cache
from trexlib.utils.string_util import random_number
from flask_login import UserMixin
import logging  
from trexlib.utils.security_util import generate_user_id, hash_password

from trexlib import conf as lib_conf
from trexmodel import conf as model_conf
import json
from json import JSONEncoder

logger = logging.getLogger("model")

class UserMin(BaseNModel, DictModel, UserMixin):
    
    #system internal usage
    user_id                     = ndb.StringProperty(required=True)
    password                    = ndb.StringProperty(required=False)
    gravatar_url                = ndb.StringProperty(required=False)
    
    #---------------------------------------------------------------------------
    # User Personal Details
    #---------------------------------------------------------------------------
    name                        = ndb.StringProperty(required=False)
    reset_password_reminder     = ndb.BooleanProperty(required=False, default=False)
    
    created_datetime            = ndb.DateTimeProperty(required=False, auto_now_add=True)
    last_login_datetime         = ndb.DateTimeProperty(required=False)

    locked                      = ndb.BooleanProperty(required=False, default=False)
    active                      = ndb.BooleanProperty(required=True, default=True)
    try_count                   = ndb.IntegerProperty(required=False)
    
    dict_properties             = ['user_id', 'name', 'email', 'gravatar_url', 'active', 'locked']
    
    @classmethod
    #@cache.memoize(timeout=60)
    def get_by_user_id(cls, user_id):
        logger.debug('UserMin.get_by_user_id: read from database')
        return cls.query(cls.user_id==user_id).get()
    
    def get_id(self):
        return self.user_id
    
    @property
    def is_super_user(self):
        return False
    
    @property
    def is_admin_user(self):
        return False
    
class Role(BaseNModel):
    id              = ndb.StringProperty(required=True)
    name            = ndb.StringProperty(required=True)
    description     = ndb.TextProperty(required=False)


class UserBase(UserMin):
    #---------------------------------------------------------------------------
    # User System Generated fields
    #---------------------------------------------------------------------------
    reference_code              = ndb.StringProperty(required=True)
    
    #---------------------------------------------------------------------------
    # User Mutual Mandatory fields
    #---------------------------------------------------------------------------
    email                       = ndb.StringProperty(required=False)
    mobile_phone                = ndb.StringProperty(required=False)
    
    #---------------------------------------------------------------------------
    # User Personal Details
    #---------------------------------------------------------------------------
    birth_date                  = ndb.DateProperty(required=False, indexed=False) 
    birth_date_date_str         = ndb.StringProperty(required=False)
    gender                      = ndb.StringProperty(required=False, choices=set([
                                                                    model_conf.GENDER_MALE_CODE, 
                                                                    model_conf.GENDER_FEMALE_CODE
                                                                    ]))
    national_id                 = ndb.StringProperty(required=False, )
    
    #---------------------------------------------------------------------------
    # User Contact Details
    #---------------------------------------------------------------------------
    
    country                     = ndb.StringProperty(required=False, default=lib_conf.DEFAULT_COUNTRY_CODE)
    state                       = ndb.StringProperty(required=False)
    city                        = ndb.StringProperty(required=False, )
    postcode                    = ndb.StringProperty(required=False, )
    address                     = ndb.StringProperty(required=False, )
    
    #---------------------------------------------------------------------------
    # User account activation required fields
    #---------------------------------------------------------------------------
    is_email_verified           = ndb.BooleanProperty(required=False, default=False)
    is_mobile_phone_verified    = ndb.BooleanProperty(required=False, default=False)

    modified_datetime           = ndb.DateTimeProperty(required=True, auto_now=True)
    
    @property
    def age(self):
        if self.birth_date:
            from dateutil.relativedelta import relativedelta
            from datetime import date
            today   = date.today()
            __age   = relativedelta(today, self.birth_date)
            return __age.years
        else:
            return 0
    
    @classmethod
    def get_by_email(cls, email):
        return User.query(ndb.AND(User.email==email)).get()
        
    @classmethod
    def get_by_mobile_phone(cls, mobile_phone):
        return User.query(ndb.AND(User.mobile_phone==mobile_phone)).get()
    
    @classmethod
    def create(cls, name=None, email=None, mobile_phone=None, 
               gender=None, birth_date=None, 
               password=None):
        
        user_id             = generate_user_id()
        reference_code      = random_number(16) 
        birth_date_date_str = None
        
        if birth_date:
            birth_date_date_str = birth_date.strftime('%d/%m')
        
        created_user = cls(user_id=user_id, name=name, email=email, mobile_phone=mobile_phone, 
                           gender=gender, birth_date=birth_date, birth_date_date_str=birth_date_date_str, 
                           reference_code=reference_code
                           )
        
        hashed_password = hash_password(user_id, password)
        created_user.password = hashed_password
            
        created_user.put()
        
        return created_user
    
    @classmethod
    def update_contact(cls, user_acct, address=None, city=None, state=None, postcode=None, country=None):
        user_acct.address    = address
        user_acct.city       = city
        user_acct.state      = state
        user_acct.postcode   = postcode
        user_acct.country    = country
        user_acct.put()
        
    @classmethod
    def update_biodata(cls, user_acct, gender=None, birth_date=None):
        user_acct.gender        = gender
        user_acct.birth_date    = birth_date
        birth_date_date_str     = None
        if birth_date:
            birth_date_date_str = birth_date.strftime('%d/%m')
        
        user_acct.birth_date_date_str    = birth_date_date_str
        
        user_acct.put()    
        
    
    
class User(UserBase):
    redeem_pin               = ndb.StringProperty(required=False)
    
    dict_properties            = [
                                    'mobile_phone', 'email', 'password', 'name', 'birth_date', 'gender',
                                    'reference_code', 'country', 'state', 'city', 'postcode', 
                                    'address', 'redeem_pin'

                                    ]

    #unique_attributes = 'email,username'

    audit_properties            = [
                                    'mobile_phone', 'email', 'password', 'name', 'birth_date', 'gender',
                                    'reference_code', 'country', 'state', 'city', 'postcode', 
                                    'address', 'redeem_pin'

                                    ]

    unicode_properties = ['name', 'address', 'city']

    export_properties = (
        ('User Id','user_id'),
        ('Name','name'),
        ('Reference Code','reference_code'),
        ('Mobile Phone','mobile_phone'),
        ('Email','email'),
        ('Gender','gender'),
        ('Date of Birth','birth_date'),
        ('Country','country_desc'),
        ('State','state_desc'),
        ('City','city'),
        ('Address','address'),
        ('Registered Datetime','created_datetime'),

    )

    def __repr__(self):
        return '''
                User[key=%s, 
                name=%s, 
                email=%s, 
                mobile_phone=%s, 
                country=%s, 
                locked=%s,
                active=%s
                ]
                ''' % (self.key_in_str, self.name, self.email, self.mobile_phone, self.country, self.locked, self.active)

    
        
    @property
    def is_active(self):
        """Returns `True` if the user is active."""
        logger.info('calling is_active')
        return self.active

      
class LoggedInUser(UserMixin, DictModel):
    
    def __init__(self, json_object, is_super_user=False, is_admin_user=False, is_merchant_user=False):
        
        logging.debug('json_object=%s', json_object)
        
        self.user_id            = json_object.get('user_id') 
        self.name               = json_object.get('name')
        self.email              = json_object.get('email')
        self.gravatar_url       = json_object.get('gravatar_url')
        self.active             = json_object.get('active')
        self.locked             = json_object.get('locked')
        self.permission         = json_object.get('permission')
        self.is_super_user      = json_object.get('is_super_user') or is_super_user
        self.is_admin_user      = json_object.get('is_admin_user') or is_admin_user
        self.is_merchant_user   = json_object.get('is_merchant_user') or is_merchant_user
        self.permission         = json_object.get('permission')
        
        self.show_key_in_dict   = False
        
        self.dict_properties     = ['user_id', 'name', 'gravatar_url', 'active', 'locked', 
                                    'is_super_user', 'is_admin_user', 'is_merchant_user', 'permission']
        
    @property
    def is_authenticated(self):
        return True
    
    @property
    def is_active(self):
        return self.active
    
    @property
    def is_anonymous(self):
        return False    

    



'''
Created on 13 Apr 2020

@author: jacklok
'''
from google.cloud import ndb
import logging
from trexmodel import conf as model_conf
from trexmodel.models.datastore.ndb_models import BaseNModel

 
class Todo(BaseNModel):
    
    name                    = ndb.StringProperty(required=True)
    desc                    = ndb.TextProperty(required=False, indexed=False)
    created_datetime        = ndb.DateTimeProperty(auto_now_add=True)
    modified_datetime       = ndb.DateTimeProperty(auto_now=True)
    
    dict_properties         = ['name', 'desc']
    dict_full_properties    = ['key','name', 'desc', 'created_datetime', 'modified_datetime']
    
    @staticmethod
    def create(name, desc=None):
        return Todo(name=name, desc=desc).put()
    
    @staticmethod
    def update(todo_key, name=None, desc=None):
        todo =  Todo.get(todo_key)
        if todo:
            todo.name = name
            todo.desc = desc
            todo.put()
    
    @staticmethod
    def list():
        return Todo.query().fetch(limit=model_conf.MAX_FETCH_RECORD)
    
    @staticmethod
    def get(todo_key_str):
        todo_key = ndb.Key(urlsafe=todo_key_str)
        return todo_key.get()
    
    @staticmethod
    def delete(todo_key_str):
        todo_key = ndb.Key(urlsafe=todo_key_str)
        
        todo_key.delete()
        
      


import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
     name='trex-model',  
     version='0.1.39',
     author="Jack Lok",
     author_email="sglok77@gmail.com",
     description="TRex database module package",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://bitbucket.org/lokjac/trex-model",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
     include_package_data = True,
     install_requires=[            
          'google-cloud-firestore',
          'google_cloud_datastore==1.11.0',
          'google-cloud-ndb',
          'six',
          'trex-lib',
          'flask-login',
          'orderedset',
      ]
 )




'''
Created on 19 Feb 2021

@author: jacklok
'''
from orderedset import OrderedSet

REWARD_BASE_ON_VISIT            = 'visit'
REWARD_BASE_ON_SPENDING         = 'spending'
REWARD_BASE_ON_PREPAID          = 'prepaid'
REWARD_BASE_ON_MEMBER           = 'membership'
REWARD_BASE_ON_REFER            = 'refer'
REWARD_BASE_ON_GIVEAWAY         = 'giveaway'

REWARD_BASE_SET                 = (REWARD_BASE_ON_VISIT, REWARD_BASE_ON_SPENDING, REWARD_BASE_ON_PREPAID, REWARD_BASE_ON_GIVEAWAY, REWARD_BASE_ON_REFER)

REWARD_FORMAT_POINT             = 'point'
REWARD_FORMAT_STAMP             = 'stamp'
REWARD_FORMAT_CASH              = 'cash'
REWARD_FORMAT_PREPAID_CASH      = 'cash'
REWARD_FORMAT_VOUCHER           = 'voucher'
REWARD_FORMAT_DISCOUNT          = 'discount'
REWARD_FORMAT_MIXED             = 'mixed'

REWARD_FORMAT_SET               = (REWARD_FORMAT_POINT, REWARD_FORMAT_STAMP, REWARD_FORMAT_CASH, REWARD_FORMAT_VOUCHER, REWARD_FORMAT_MIXED, REWARD_FORMAT_PREPAID_CASH)

PROGRAM_STATUS_REWARD_BASE          = 'reward_base'
PROGRAM_STATUS_REWARD_FORMAT        = 'reward_format'
PROGRAM_STATUS_REWARD_SCHEME        = 'reward_scheme'
PROGRAM_STATUS_REWARD_DETAILS       = 'reward_details'
PROGRAM_STATUS_REWARD_EXCLUSIVITY   = 'exclusivity'
PROGRAM_STATUS_REVIEW               = 'review'
PROGRAM_STATUS_PUBLISH              = 'published'

PROGRAM_REWARD_GIVEAWAY_METHOD_AUTO     = 'auto'
PROGRAM_REWARD_GIVEAWAY_METHOD_MANUAL   = 'manual'

PROGRAM_STATUS                  = OrderedSet([PROGRAM_STATUS_REWARD_BASE, 
                                              PROGRAM_STATUS_REWARD_SCHEME, 
                                              PROGRAM_STATUS_REWARD_EXCLUSIVITY,
                                              PROGRAM_STATUS_PUBLISH
                                              ])


VOUCHER_STATUS_BASE                 = 'voucher_base'
VOUCHER_STATUS_CONFIGURATION        = 'voucher_configuration'
VOUCHER_STATUS_UPLOAD_MATERIAL      = 'upload_material'
VOUCHER_STATUS_PUBLISH              = 'published'

VOUCHER_STATUS                  = OrderedSet([VOUCHER_STATUS_BASE, 
                                              VOUCHER_STATUS_CONFIGURATION,
                                              VOUCHER_STATUS_UPLOAD_MATERIAL, 
                                              VOUCHER_STATUS_PUBLISH
                                              ])

PRORRAM_NEXT_STEP_AND_COMPLELTED_STATUS_MAPPING = {
                                                    PROGRAM_STATUS_REWARD_BASE: 2
                                                    }


REWARD_EFFECTIVE_TYPE_IMMEDIATE         = 'immediate'
REWARD_EFFECTIVE_TYPE_SPECIFIC_DATE     = 'date'
REWARD_EFFECTIVE_TYPE_AFTER_MONTH       = 'month'
REWARD_EFFECTIVE_TYPE_AFTER_WEEK        = 'week'
REWARD_EFFECTIVE_TYPE_AFTER_DAY         = 'day'


REWARD_EXPIRATION_TYPE_SPECIFIC_DATE     = 'date'
REWARD_EXPIRATION_TYPE_AFTER_YEAR        = 'year'
REWARD_EXPIRATION_TYPE_AFTER_MONTH       = 'month'
REWARD_EXPIRATION_TYPE_AFTER_WEEK        = 'week'
REWARD_EXPIRATION_TYPE_AFTER_DAY         = 'day'


MEMBERSHIP_EXPIRATION_TYPE_AFTER_YEAR       = 'year'
MEMBERSHIP_EXPIRATION_TYPE_AFTER_MONTH      = 'month'
MEMBERSHIP_EXPIRATION_TYPE_AFTER_WEEK       = 'week'
MEMBERSHIP_EXPIRATION_TYPE_SPECIFIC_DATE    = 'date'
MEMBERSHIP_EXPIRATION_TYPE_NO_EXPIRY        = 'no_expiry'

MEMBERSHIP_EFFECTIVE_TYPE_IMMEDIATE         = 'immediate'
MEMBERSHIP_EFFECTIVE_TYPE_SPECIFIC_DATE     = 'date'
MEMBERSHIP_EFFECTIVE_TYPE_AFTER_DAY         = 'day'

MEMBERSHIP_ENTIEL_QUALIFICATION_TYPE_AUTO_ASSIGN                    = 'auto_assign'
MEMBERSHIP_ENTIEL_QUALIFICATION_TYPE_ACCUMULATED_SPENDING_AMOUNT    = 'acc_spending'
MEMBERSHIP_ENTIEL_QUALIFICATION_TYPE_ACCUMULATED_POINT_AMOUNT       = 'acc_point'
MEMBERSHIP_ENTIEL_QUALIFICATION_TYPE_ACCUMULATED_STAMP_AMOUNT       = 'acc_stamp'
MEMBERSHIP_ENTIEL_QUALIFICATION_TYPE_EXCEED_SPENDING_AMOUNT         = 'exceed_spending'
MEMBERSHIP_ENTIEL_QUALIFICATION_TYPE_EXCEED_RELOAD_AMOUNT           = 'exceed_reload'

MEMBERSHIP_UPGRADE_EXPIRY_TYPE_CONTINUE_EXPIRY                      = 'cont_expiry'
MEMBERSHIP_UPGRADE_EXPIRY_TYPE_NEW_EXPIRY                           = 'new_expiry'


REWARD_STATUS_VALID        = 'valid'
REWARD_STATUS_REDEEMED     = 'redeemed'
REWARD_STATUS_REVERTED     = 'reverted'

REDEEM_CODE_LENGTH                                                  = 12

def get_program_completed_status_index(completed_status):
    return PROGRAM_STATUS.index(completed_status)

def get_voucher_completed_status_index(completed_status):
    return VOUCHER_STATUS.index(completed_status) 

def is_program_current_status_reach(checking_status, completed_status):
    completed_status_index  = get_program_completed_status_index(completed_status)
    checking_status_index   = get_program_completed_status_index(checking_status)
    
    print('completed_status_index=%s'%completed_status_index)
    print('checking_status_index=%s'%checking_status_index)
    
    return checking_status_index<=completed_status_index+1

def is_voucher_current_status_reach(checking_status, completed_status):
    completed_status_index  = get_voucher_completed_status_index(completed_status)
    checking_status_index   = get_voucher_completed_status_index(checking_status)
    
    print('completed_status_index=%s'%completed_status_index)
    print('checking_status_index=%s'%checking_status_index)
    
    return checking_status_index<=completed_status_index+1

def is_valid_to_update_program_status(checking_status, completed_status):
    completed_status_index  = get_program_completed_status_index(completed_status)
    checking_status_index   = get_program_completed_status_index(checking_status)
    
    return checking_status_index<=completed_status_index+1

def is_valid_to_update_voucher_status(checking_status, completed_status):
    completed_status_index  = get_voucher_completed_status_index(completed_status)
    checking_status_index   = get_voucher_completed_status_index(checking_status)
    
    return checking_status_index<=completed_status_index+1


def is_existing_program_status_higher_than_updating_status(checking_status, completed_status):
    completed_status_index  = get_program_completed_status_index(completed_status)
    checking_status_index   = get_program_completed_status_index(checking_status)
    
    return checking_status_index<completed_status_index

def is_existing_voucher_status_higher_than_updating_status(checking_status, completed_status):
    completed_status_index  = get_voucher_completed_status_index(completed_status)
    checking_status_index   = get_voucher_completed_status_index(checking_status)
    
    return checking_status_index<completed_status_index

def is_existing_program_status_final_state(completed_status):
    completed_status_index  = get_program_completed_status_index(completed_status)
    
    return completed_status_index==len(PROGRAM_STATUS)-2

def is_existing_voucher_status_final_state(completed_status):
    completed_status_index  = get_voucher_completed_status_index(completed_status)
    
    return completed_status_index==len(PROGRAM_STATUS)-2

def program_completed_progress_percentage(completed_status):
    completed_status_index  = get_program_completed_status_index(completed_status)
    print('program_completed_progress_percentage: completed_status_index(%s)=%s'% (completed_status,completed_status_index))
    return int((completed_status_index+1)/len(PROGRAM_STATUS) * 100)    
    

